from collections import namedtuple, OrderedDict
from itertools import chain
from os import makedirs, rename, scandir, listdir
from os.path import (join as p, exists, relpath, isdir, isfile,
        expanduser, expandvars, realpath)
from struct import pack
import errno
import hashlib
import json
import logging
import re
import shutil
import io

from rdflib import plugin
from rdflib.parser import Parser, create_input_source
from rdflib.term import URIRef
import six
from textwrap import dedent
import transaction
import yaml

from .. import OWMETA_PROFILE_DIR, connect
from ..context import DEFAULT_CONTEXT_KEY, IMPORTS_CONTEXT_KEY, Context
from ..mapper import CLASS_REGISTRY_CONTEXT_KEY, CLASS_REGISTRY_CONTEXT_LIST_KEY, Mapper
from ..context_common import CONTEXT_IMPORTS
from ..data import Data
from ..file_match import match_files
from ..file_lock import lock_file
from ..file_utils import hash_file
from ..graph_serialization import write_canonical_to_file
from ..rdf_utils import BatchAddGraph
from ..utils import FCN, aslist

from .archive import Unarchiver
from .common import (find_bundle_directory, fmt_bundle_directory, BUNDLE_MANIFEST_FILE_NAME,
                     BUNDLE_INDEXED_DB_NAME, validate_manifest, BUNDLE_MANIFEST_VERSION)
from .exceptions import (NotADescriptor, BundleNotFound, NoRemoteAvailable, NoBundleLoader,
                         NotABundlePath, MalformedBundle, NoAcceptableUploaders,
                         FetchTargetIsNotEmpty, TargetIsNotEmpty, UncoveredImports,
                         CircularDependencyDetected)
from .loaders import LOADER_CLASSES, UPLOADER_CLASSES, load_entry_point_loaders

from urllib.parse import quote as urlquote, unquote as urlunquote


L = logging.getLogger(__name__)

DEFAULT_BUNDLES_DIRECTORY = p(OWMETA_PROFILE_DIR, 'bundles')
'''
Default directory for the bundle cache
'''

DEFAULT_REMOTES_DIRECTORY = p(OWMETA_PROFILE_DIR, 'remotes')
'''
Default directory for descriptors of user-level remotes as opposed to project-specific
remotes
'''


class Remote(object):
    '''
    A place where bundles come from and go to
    '''
    def __init__(self, name, accessor_configs=()):
        '''
        Parameters
        ----------
        name : str
            The name of the remote
        accessor_configs : iterable of AccessorConfig
            Configs for how you access the remote
        '''

        self.name = name
        ''' Name of the remote '''

        self.accessor_configs = list(accessor_configs)
        '''
        Configs for how you access the remote.

        One might configure mirrors or replicas for a given bundle repository as multiple
        accessor configs
        '''

        self.file_name = None
        '''
        If read from a file, the remote should have this attribute set to its source
        file's path
        '''

    def add_config(self, accessor_config):
        '''
        Add the given accessor config to this remote

        Parameters
        ----------
        accessor_config : AccessorConfig
            The config to add

        Returns
        -------
        bool
            `True` if the accessor config was added (meaning there's no equivalent one
            already set for this remote). Otherwise, `False`.
        '''
        if accessor_config in self.accessor_configs:
            return False
        self.accessor_configs.append(accessor_config)
        return True

    def generate_loaders(self):
        '''
        Generate the bundle loaders for this remote.

        Loaders are generated from `accessor_configs` and `LOADER_CLASSES` according with
        which type of `.Loader` can load a type of accessor
        '''
        for ac in self.accessor_configs:
            for lc in LOADER_CLASSES:
                if lc.can_load_from(ac):
                    loader = lc(ac)
                    yield loader

    def generate_uploaders(self):
        '''
        Generate the bundle uploaders for this remote
        '''
        for ac in self.accessor_configs:
            for uc in UPLOADER_CLASSES:
                if uc.can_upload_to(ac):
                    loader = uc(ac)
                    yield loader

    def write(self, out):
        '''
        Serialize the `Remote` and write to `out`

        Parameters
        ----------
        out : :term:`file object`
            Target for writing the remote
        '''
        yaml.dump(self, out)

    @classmethod
    def read(cls, inp):
        '''
        Read a serialized `Remote`

        Parameters
        ----------
        inp : :term:`file object`
            File-like object containing the serialized `Remote`
        '''
        res = yaml.unsafe_load(inp)
        assert isinstance(res, cls)
        return res

    def __eq__(self, other):
        return (self.name == other.name and
                self.accessor_configs == other.accessor_configs)

    def __hash__(self):
        return hash((self.name, self.accessor_configs))

    def __str__(self):
        if self.accessor_configs:
            accessors = '\n' + '\n'.join('    ' + '\n    '.join(str(acc).split('\n')) for acc in self.accessor_configs)
        else:
            accessors = ' <none>'
        return dedent('''\
        {name}
        Accessors:{accessors}''').format(name=self.name,
            accessors=accessors)

    def __repr__(self):
        return f'{FCN(type(self))}({self.name!r}, {self.accessor_configs!r})'


class DependencyDescriptor(namedtuple('_DependencyDescriptor',
        ('id', 'version', 'excludes'))):
    __slots__ = ()

    def __new__(cls, id, version=None, excludes=()):
        return super(DependencyDescriptor, cls).__new__(cls, id, version, excludes)


class AccessorConfig(object):
    '''
    Configuration for accessing a `Remote`. `Loaders <Loader>` are added to a remote according to
    which accessors are available
    '''

    def __eq__(self, other):
        raise NotImplementedError()

    def __hash__(self):
        raise NotImplementedError()


class _DepList(list):
    def add(self, dd):
        self.append(dd)


class URLConfig(AccessorConfig):
    '''
    Configuration for accessing a remote with just a URL.

    Note that URLConfigs should be pickle-able since they are written to a YAML file as
    part of the `.Remote` they're apart of.
    '''

    def __init__(self, url):
        self.url = url

    def __eq__(self, other):
        return isinstance(other, URLConfig) and self.url == other.url

    def __hash__(self):
        return hash(self.url)

    def __str__(self):
        return '{}(url={})'.format(FCN(type(self)), repr(self.url))

    @classmethod
    def register(cls, scheme):
        URL_CONFIG_MAP[scheme] = cls

    __repr__ = __str__


URL_CONFIG_MAP = {}
'''
`URLConfigs <URLConfig>` by scheme. Can be populated by pkg_resources entry points
'''


class Descriptor(object):
    '''
    Descriptor for a bundle.

    The descriptor is sufficient to build a distributable bundle directory tree from a
    `~rdflib.graph.ConjunctiveGraph` and a set of files (see `Installer`).
    '''
    def __init__(self, ident, **kwargs):
        self.id = ident
        self._set(kwargs)

    @classmethod
    def make(cls, obj):
        '''
        Makes a descriptor from the given object.

        Parameters
        ----------
        obj : a `dict-like object <dict>`
            An object with parameters for the Descriptor. Typically a dict

        Returns
        -------
        Descriptor
            The created descriptor
        '''
        res = cls(ident=obj['id'])
        res._set(obj)
        return res

    @classmethod
    def load(cls, descriptor_source):
        '''
        Load a descriptor from a YAML record

        Parameters
        ----------
        descriptor_source : str or :term:`file object`
            The descriptor source. Handled by `yaml.safe_load
            <https://pyyaml.org/wiki/PyYAMLDocumentation#the-yaml-package>`_

        Raises
        ------
        .NotADescriptor
            Thrown when the object loaded from `descriptor_source` isn't a `dict`
        '''
        dat = yaml.safe_load(descriptor_source)
        if isinstance(dat, dict):
            return cls.make(dat)
        else:
            raise NotADescriptor()

    def dump(self, output):
        '''
        Save a descriptor to a file as a YAML record

        Parameters
        ----------
        output : :term:`file object`
            The file to save the descriptor to
        '''
        dct = dict()
        dct['id'] = self.id
        dct['name'] = self.name
        dct['version'] = self.version
        dct['description'] = self.description
        dct['patterns'] = sorted(('rgx:' if isinstance(RegexURIPattern, p) else '') + p._pattern
                for p in self.patterns)
        dct['includes'] = sorted(str(inc.include) for inc in self.includes)
        # XXX: Test this
        dct['includes'].extend({inc: {'empty': True}} for inc in sorted(self.empties))
        dct['dependencies'] = [dict(d._asdict()) for d in self.dependencies]
        if self.files is not None:
            dct['files'] = self.files.to_dict()
        yaml.dump(dct, output)

    def _set(self, obj):
        self.name = obj.get('name', self.id)
        self.version = obj.get('version', 1)
        self.description = obj.get('description', None)
        self.patterns = set(make_pattern(x) for x in obj.get('patterns', ()))
        self.includes = set(make_include_func(x) for x in obj.get('includes', ()))
        self.empties = {uri for uri, options in (inc.popitem()
            for inc in obj.get('includes', ())
            if isinstance(inc, dict))
            if options.get('empty', False) is True}

        deps_set = set()
        deps = _DepList()
        for x in obj.get('dependencies', ()):
            if isinstance(x, six.string_types):
                dd = DependencyDescriptor(x)
            elif isinstance(x, dict):
                dd = DependencyDescriptor(**x)
            else:
                dd = DependencyDescriptor(*x)
            if dd not in deps_set:
                deps.append(dd)
                deps_set.add(dd)
        self.dependencies = deps
        self.files = FilesDescriptor.make(obj.get('files', None))

    def __str__(self):
        return (FCN(type(self)) + '(ident={},'
                'name={},version={},description={},'
                'patterns={},includes={},'
                'files={},dependencies={})').format(
                        repr(self.id),
                        repr(self.name),
                        repr(self.version),
                        repr(self.description),
                        repr(self.patterns),
                        repr(self.includes),
                        repr(self.files),
                        repr(self.dependencies))


class Bundle:
    '''
    Main entry point for using bundles

    Typical usage is something like this::

        >>> with Bundle('example/bundleId', version=42) as bnd:
        ...     for aDataObject in bnd(DataObject)().load():
        ...         # Do something with `aDataObject`
        ...         print(aDataObject)
        DataObject(<http://example.org/entities#aDataObject>)
    '''

    def __init__(self, ident, bundles_directory=DEFAULT_BUNDLES_DIRECTORY, version=None,
            conf=None, remotes=None, remotes_directory=DEFAULT_REMOTES_DIRECTORY,
            transaction_manager=None):
        '''
        .. note::

            Paths, `bundles_directory` and `remotes_directory`, will have symbolic links,
            environment variables, and "~" (for the current user's home directory)
            expanded when the `Bundle` is initialized. To reflect changes to symbolic
            links or home directories, the `bundles_directory` or `remotes_directory`
            attributes must be updated directly or a new instance must be created.

        Parameters
        ----------
        ident : str
            Bundle ID
        bundles_directory : str, optional
            Path to the bundles directory. Defaults to `.DEFAULT_BUNDLES_DIRECTORY`
        version : int, optional
            Bundle version to access. By default, the latest version will be used.
        conf : .Configuration or dict, optional
            Configuration to add to the one created for the bundle automatically. Values
            for the default imports context (`.IMPORTS_CONTEXT_KEY`), the default context
            (`.DEFAULT_CONTEXT_KEY`) and store (``'rdf.store'``, ``'rdf.source'``, and,
            ``'rdf.store_conf'``) will be ignored and overwritten.
        remotes : iterable of Remote or str, optional
            A subset of remotes and additional remotes to fetch from. See `Fetcher.fetch`
        remotes_directory : str, optional
            The directory to load `Remotes <Remote>` from in case a bundle is not in the
            bundle cache. Defaults to `.DEFAULT_REMOTES_DIRECTORY`
        transaction_manager : `transaction.TransactionManager`, optional
            Transaction manager
        '''
        if not ident or not isinstance(ident, str):
            raise ValueError('ident must be a non-empty string')
        self.ident = ident
        if not bundles_directory:
            bundles_directory = DEFAULT_BUNDLES_DIRECTORY
        self.bundles_directory = realpath(expandvars(expanduser(bundles_directory)))
        if not conf:
            conf = {}

        conf.update({'rdf.source': 'default'})
        self.version = version
        self.remotes = remotes
        # XXX: Look at how we bring in projects remotes directory
        if not remotes_directory:
            remotes_directory = DEFAULT_REMOTES_DIRECTORY
        self.remotes_directory = realpath(expandvars(expanduser(remotes_directory)))

        if transaction_manager is not None:
            self.transaction_manager = transaction_manager
        else:
            self.transaction_manager = BundleTransactionManager(explicit=True)

        self._store_config_builder = \
            BundleDependentStoreConfigBuilder(
                    bundles_directory=bundles_directory,
                    remotes_directory=remotes_directory,
                    remotes=remotes,
                    transaction_manager=self.transaction_manager)

        self._bundle_dep_mgr = BundleDependencyManager(
                bundles_directory=self.bundles_directory,
                remotes=self.remotes,
                remotes_directory=self.remotes_directory,
                dependencies=self.dependencies)
        self._given_conf = conf
        self.conf = None
        self._contexts = None
        self.connection = None
        ''' The owmeta_core connection to the bundle's indexed database '''

        self._bundle_context = None
        self._loaded_dependencies = dict()

    @property
    def identifier(self):
        return self.ident

    def resolve(self):
        try:
            bundle_directory = self._get_bundle_directory()
        except BundleNotFound:
            bundle_directory = self._fetch_bundle(self.ident, self.version)
        return bundle_directory

    @property
    def manifest_data(self):
        bundle_directory = self.resolve()
        with open(p(bundle_directory, BUNDLE_MANIFEST_FILE_NAME)) as mf:
            return json.load(mf)

    def _get_bundle_directory(self):
        # - look up the bundle in the bundle cache
        # - generate a config based on the current config load the config
        # - make a database from the graphs, if necessary (similar to `owm regendb`). If
        #   delete the existing database if it doesn't match the store config
        return find_bundle_directory(self.bundles_directory, self.ident, self.version)

    def initdb(self):
        '''
        Initialize the bundle's `conf` `~owmeta_core.data.Data` instance
        '''
        if self.conf is None:
            bundle_directory = self.resolve()
            self.conf = Data().copy(self._given_conf)
            with open(p(bundle_directory, BUNDLE_MANIFEST_FILE_NAME)) as mf:
                manifest_data = json.load(mf)
                self.conf[DEFAULT_CONTEXT_KEY] = manifest_data.get(DEFAULT_CONTEXT_KEY)
                self.conf[IMPORTS_CONTEXT_KEY] = manifest_data.get(IMPORTS_CONTEXT_KEY)
                self.conf[CLASS_REGISTRY_CONTEXT_KEY] = manifest_data.get(CLASS_REGISTRY_CONTEXT_KEY)
                if CLASS_REGISTRY_CONTEXT_LIST_KEY not in self.conf:
                    crctx_ids = []
                    for dep in self._bundle_dep_mgr.load_dependencies_transitive():
                        crctx_id = dep.manifest_data.get(CLASS_REGISTRY_CONTEXT_KEY)
                        if crctx_id:
                            crctx_ids.append(crctx_id)
                    self.conf[CLASS_REGISTRY_CONTEXT_LIST_KEY] = crctx_ids
            indexed_db_path = p(bundle_directory, BUNDLE_INDEXED_DB_NAME)
            store_name, store_conf = self._store_config_builder.build(
                    indexed_db_path,
                    manifest_data.get('dependencies', ()))
            self.conf['rdf.store'] = store_name
            self.conf['rdf.store_conf'] = store_conf
            self.connection = connect(conf=self.conf)

    def _fetch_bundle(self, bundle_ident, version):
        remotes_list = list(retrieve_remotes(self.remotes_directory))
        f = Fetcher(self.bundles_directory, remotes_list)
        return f.fetch(bundle_ident, version, self.remotes)

    @property
    def contexts(self):
        '''
        `List <list>` of `str`. Context IDs in this bundle
        '''
        # Since bundles are meant to be immutable, we won't need to add
        if self._contexts is not None:
            return self._contexts
        bundle_directory = self.resolve()
        contexts = list()
        graphs_directory = p(bundle_directory, 'graphs')
        idx_fname = p(graphs_directory, 'index')
        if not exists(idx_fname):
            raise Exception('Cannot find an index at {}'.format(repr(idx_fname)))
        with open(idx_fname, 'rb') as index_file:
            for l in index_file:
                l = l.strip()
                if not l:
                    continue
                ctx, _ = l.split(b'\x00')
                contexts.append(ctx.decode('UTF-8'))
        self._contexts = frozenset(contexts)
        return self._contexts

    @property
    def rdf(self):
        self.initdb()
        return self.conf['rdf.graph']

    def __str__(self):
        return f'Bundle({self.ident}' + (')' if self.version is None else f', {self.version})')

    def __enter__(self):
        self.initdb()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Close the database connection
        self.connection.disconnect()
        self.connection = None
        self.conf = None

    def dependencies(self):
        return self.manifest_data.get('dependencies', ())

    def load_dependencies_transitive(self):
        '''
        Load dependencies from this bundle transitively

        Yields
        ------
        Bundle
            A direct or indirect dependency of this bundle
        '''
        return self._bundle_dep_mgr.load_dependencies_transitive()

    def load_dependencies(self):
        '''
        Load direct dependencies of this bundle

        Yields
        ------
        Bundle
            A direct dependency of this bundle
        '''
        return self._bundle_dep_mgr.load_dependencies()

    def _lookup_context_bundle(self, context_id):
        owner = self._bundle_dep_mgr.lookup_context_bundle(
                self.contexts,
                context_id)
        if owner is self._bundle_dep_mgr:
            return self

    def _load_dependency(self, dependencies_item):
        try:
            return self._bundle_dep_mgr._load_dependency(dependencies_item)
        except BundleDependencyConfigIsMalformed as e:
            bundle_directory = self.resolve()
            raise MalformedBundle(bundle_directory, str(e)) from e

    def __call__(self, target):
        if not target or not hasattr(target, 'contextualize'):
            return target

        self.initdb()
        if self._bundle_context is None:
            self._bundle_context = _BundleContext(
                    None, conf=self.conf, bundle=self).stored

        return target.contextualize(self._bundle_context)


class BundleDependencyManager(object):
    '''
    Finds the bundle in which a context is defined.

    For a given bundle graph, that there is *one* Bundle that "owns" a given context.
    Although multiple bundles may provide that context, the one closest to the root of the
    graph which provides some statements in that context is called the owner. Note that
    this does not mean that bundles on which the owner depends do not also be queried;
    however, the exact behavior is up to the component that uses this component.
    '''

    def __init__(self, dependencies, **common_bundle_arguments):
        '''
        Parameters
        ----------
        dependencies : function
            Function that returns a sequence of dependency descriptors
        '''
        self._loaded_dependencies = dict()
        self._common_bundle_arguments = common_bundle_arguments
        self.dependencies = dependencies

    def load_dependencies_transitive(self):
        '''
        Load dependencies from this bundle transitively.

        Any given version of a bundle will be yielded at most once regardless of how many
        times that version of the bundle appears in the dependency graph. Dependencies
        will yielded in topological sort order, so every dependency a Bundle declares will
        be yielded before any of its transitive dependencies.

        Yields
        ------
        Bundle
            A direct or indirect dependency of this bundle
        '''
        border = {None: self}
        dependants = OrderedDict()
        seen = dict()
        while border:
            # allegedly, OrderedDict isn't optimized for iteration speed, but better to
            # signal that order is important in how the border is iterated over.
            new_border = OrderedDict()
            for bnd in border.values():
                bkey = None if bnd is self else (bnd.ident, bnd.version)
                for d_bnd in bnd.load_dependencies():
                    key = (d_bnd.ident, d_bnd.version)

                    this_dependants = dependants.get(key)

                    if this_dependants is None:
                        if bkey is None:
                            dependants[key] = set()
                        else:
                            dependants[key] = set([bkey])
                    else:
                        this_dependants.add(bkey)

                    if key in seen:
                        continue
                    seen[key] = d_bnd
                    new_border[key] = d_bnd
            border = new_border

        while dependants:
            for key, this_dependants in dependants.items():
                if not this_dependants:
                    yield seen[key]
                    break
            else: # no break
                break
            del dependants[key]

            for this_dependants in dependants.values():
                this_dependants.discard(key)

        if dependants:
            # Handle the case that we didn't deplete the adjacency list, implying that we
            # have a cycle
            sio = io.StringIO()
            for dependency, this_dependants in dependants.items():
                for dependant in this_dependants:
                    sio.write(f' {dependant} -> {dependency}')

            raise CircularDependencyDetected(
                    f'Circular dependency detected. Links:{sio.getvalue()}')

    def lookup_context_bundle(self, contexts, context_id):
        if context_id is None or str(context_id) in contexts:
            return self
        for d in self.dependencies():
            d_excludes = frozenset(d.get('excludes', ()))
            if context_id in d_excludes:
                continue

            d_bnd = self._load_dependency(d)
            match = d_bnd._lookup_context_bundle(context_id)
            if match:
                return match
        return None

    def _load_dependencies(self):
        for d in self.dependencies():
            yield self._load_dependency(d)

    load_dependencies = _load_dependencies

    def _load_dependency(self, dependencies_item):
        d_id = dependencies_item.get('id')
        if not d_id:
            raise BundleDependencyConfigIsMalformed('Dependency entry is missing an identifier')
        d_version = dependencies_item.get('version')
        if not d_version:
            raise BundleDependencyConfigIsMalformed(f'Dependency entry for {d_id} is'
                    ' missing a version number')

        bundle = self._loaded_dependencies.get((d_id, d_version))
        if not bundle:
            bundle = Bundle(d_id, version=d_version,
                    **self._common_bundle_arguments)
            self._loaded_dependencies[(d_id, d_version)] = bundle
        return bundle


class BundleDependencyConfigIsMalformed(Exception):
    pass


class BundleDependentStoreConfigBuilder(object):
    '''
    Builds an RDFLib store configuration that depends on bundles.

    The process of building the store configuration requires traversing the graph of
    dependencies so that duplicate dependencies in the graph can be omitted. To support
    this process, this builder will fetch bundles as needed to resolve transitive
    dependencies
    '''
    def __init__(self, bundles_directory=None, remotes_directory=None, remotes=None,
            read_only=True, transaction_manager=None):
        if not bundles_directory:
            bundles_directory = DEFAULT_BUNDLES_DIRECTORY
        self.bundles_directory = realpath(expandvars(expanduser(bundles_directory)))

        if not remotes_directory:
            remotes_directory = DEFAULT_REMOTES_DIRECTORY
        self.remotes_directory = realpath(expandvars(expanduser(remotes_directory)))

        self.remotes = remotes
        self.read_only = read_only
        self.transaction_manager = transaction_manager

    def build(self, indexed_db_path, dependencies, bundle_directory=None):
        '''
        Builds the store configuration

        Parameters
        ----------
        indexed_db_path : str
            Path to the indexed database of the store that depends on the listed
            dependenices
        dependencies : list of dict
            List of dependencies' info, each entry including at least keys for 'id' and
            'version'
        bundle_directory : str, optional
            Path to the bundle directory for the dependent store, if the dependent store
            is a bundle. Used for information in an exceptional path, but not otherwise
            used

        Returns
        -------
        str
            The type of the store. This is the name used to look up the RDFLib store plugin
        object
            The configuration for the store. This is the object that will be passed to
            `rdflib.store.Store.open` to configure the store.
        '''
        return 'agg', self._construct_store_config(indexed_db_path, dependencies,
                read_only=self.read_only)

    __call__ = build

    def _construct_store_config(self, indexed_db_path, dependencies,
                                current_view_desc=None, view_descs=None, bundle_directory=None,
                                read_only=True):
        '''
        Parameters
        ----------
        indexed_db_path : str
            File path to the dependency database
        dependencies : list of dict
            List of dependencies' info, each entry including at least keys for 'id' and
            'version'
        current_view_desc : tuple, (_BDVD, (str, int)), optional
            Description of the current view determined by the path through dependant
            bundles
        view_descs : set of tuple, optional
            Describe the "views" on bundles within the dependency tree. The resulting
            aggregate store config have one subordinate config for each view.
        bundle_directory : str, optional
            The directory of the bundle the config is being built from. Will be determined
            by using `bundles_directory`, the bundle ID, and version for dependencies
        read_only : boolean, optional
            Whether the *top-level* config, which may not be a bundle, is read-only (the
            bundle dependencies are always read-only)
        '''
        if view_descs is None:
            view_descs = set()
        if current_view_desc is None:
            current_view_desc = _BDVD()
        dependency_configs = self._gather_dependency_configs(
                dependencies, current_view_desc, view_descs, bundle_directory)
        fs_store_config = dict(url=indexed_db_path, read_only=read_only,
                transaction_manager=self.transaction_manager)
        return [
            ('FileStorageZODB', fs_store_config)
        ] + dependency_configs

    @aslist
    def _gather_dependency_configs(self, dependencies, current_view_desc, view_descs, bundle_directory=None):
        for dd in dependencies:
            dep_view_desc = current_view_desc.merge_excludes(dd.get('excludes', ()))
            dep_ident = dd.get('id')
            dep_version = dd.get('version')
            if not dep_ident:
                if bundle_directory:
                    raise MalformedBundle(bundle_directory, 'bundle dependency descriptor is lacking an identifier')
                else:
                    raise ValueError('bundle dependency descriptor is lacking an identifier')
            if (dep_view_desc, (dep_ident, dep_version)) in view_descs:
                return
            view_descs.add((dep_view_desc, (dep_ident, dep_version)))
            tries = 0
            while tries < 2:
                try:
                    bundle_directory = find_bundle_directory(self.bundles_directory, dep_ident, dep_version)
                    with open(p(bundle_directory, BUNDLE_MANIFEST_FILE_NAME)) as mf:
                        manifest_data = json.load(mf)
                    break
                except (BundleNotFound, FileNotFoundError):
                    self._fetch_bundle(dep_ident, dep_version)
                    tries += 1

            # We don't want to include items in the configuration that aren't specified by
            # the dependency descriptor. Also, all of the optionals have defaults that
            # BundleDependencyStore handles itself, so we don't want to impose them here.
            addl_dep_confs = {k: v for k, v in dd.items()
                    if k in ('excludes',) and v}
            yield ('owmeta_core_bds', dict(type='agg',
                        conf=self._construct_store_config(
                            p(bundle_directory, BUNDLE_INDEXED_DB_NAME),
                            manifest_data.get('dependencies', ()),
                            dep_view_desc, view_descs, bundle_directory),
                        **addl_dep_confs))

    def _fetch_bundle(self, bundle_ident, version):
        remotes_list = list(retrieve_remotes(self.remotes_directory))
        f = Fetcher(self.bundles_directory, remotes_list)
        return f.fetch(bundle_ident, version, self.remotes)


class _BDVD(namedtuple('_BDVD', ('excludes',))):
    '''
    Bundle Dependency View Data (BDVD)

    Holds data we use in traversing bundle dependencies to describe the "view" on the
    bundle
    '''
    __slots__ = ()

    def __new__(cls, *args, excludes=(), **kwargs):
        return super(_BDVD, cls).__new__(cls, *args, excludes=excludes, **kwargs)

    def merge_excludes(self, excludes):
        return self._replace(excludes=self.excludes +
                tuple(e for e in excludes if e not in self.excludes))


class _BundleContext(Context):
    '''
    `Context` for a bundle.
    '''
    def __init__(self, *args, bundle, **kwargs):
        super().__init__(*args, **kwargs)
        self.bundle = bundle
        self._mapper = None

    @property
    def mapper(self):
        if self._mapper is None:
            self._mapper = _BundleMapper(bundle=self.bundle)
        return self._mapper


class _BundleMapper(Mapper):
    def __init__(self, bundle):
        try:
            bundle_conf = bundle.conf
        except AttributeError:
            raise Exception('Bundle connection has not been established.'
                    ' Call `initdb` or use the bundle in a context manager')
        super().__init__(name=f'{bundle.ident}' +
                (f'@{bundle.version}' if bundle.version else ''),
                conf=bundle_conf)
        self.bundle = bundle
        self._resolved_classes = dict()

    def resolve_class(self, rdf_type, context):
        prev_resolved_class = self._resolved_classes.get((rdf_type, context.identifier))
        if prev_resolved_class:
            return prev_resolved_class

        own_resolved_class = super().resolve_class(rdf_type, context)

        if own_resolved_class:
            self._resolved_classes[(rdf_type, context.identifier)] = own_resolved_class
            return own_resolved_class

        target_id = context.identifier
        target_bundle = self.bundle._lookup_context_bundle(target_id)
        deps = target_bundle.load_dependencies_transitive()
        for bnd in deps:
            crctx_id = bnd.manifest_data.get(CLASS_REGISTRY_CONTEXT_KEY, None)
            if not crctx_id:
                continue
            with bnd:
                resolved_class = bnd.connection.mapper.resolve_class(rdf_type, context)
                if resolved_class:
                    self._resolved_classes[(rdf_type, context.identifier)] = resolved_class
                    return resolved_class
        return None


class _RemoteHandlerMixin(object):
    '''
    Utility mixin for handling remotes

    The mixed-in class must have a `remotes` attribute which is a list of `Remote`
    '''

    def __init__(self, load_entry_points=True, **kwargs):
        '''
        Parameters
        ----------
        load_entry_points : bool, optional
            If `False`, then entry points will not be loaded
        '''
        super(_RemoteHandlerMixin, self).__init__(**kwargs)
        self.load_entry_points = load_entry_points

    def _get_remotes(self, remotes):
        ''''
        Get remotes

        Parameters
        ----------
        remotes : iterable of Remote or str
            A subset of names of remotes to act on and additional remotes to act on
        '''
        if self.load_entry_points:
            load_entry_point_loaders()
        instance_remotes = []
        additional_remotes = []
        if remotes:
            configured_remotes = {r.name: r for r in self.remotes}
            for r in remotes:
                if isinstance(r, six.text_type):
                    instance_remotes.append(configured_remotes.get(r))
                elif isinstance(r, Remote):
                    additional_remotes.append(r)
        else:
            instance_remotes = self.remotes
        has_remote = False
        for rem in chain(additional_remotes, instance_remotes):
            has_remote = True
            yield rem

        if not has_remote:
            raise NoRemoteAvailable(f'No valid remote from {remotes} among {self.remotes}'
                    f' for {self}')


class Fetcher(_RemoteHandlerMixin):
    '''
    Fetches bundles from `Remotes <Remote>`

    A fetcher takes a list of remotes, a bundle ID, and, optionally, a version number and
    downloads the bundle to a local directory. `Deployer` is, functionally, the dual of
    this class.
    '''

    def __init__(self, bundles_root, remotes, transaction_manager=None, **kwargs):
        '''
        Parameters
        ----------
        bundles_root : str
            The root directory of the bundle cache
        remotes : list of Remote or str
            List of pre-configured remotes used in calls to `fetch`
        transaction_manager : `transaction.TransactionManager`
            Transaction manager to use when populating the indexed database after fetching
        '''
        super(Fetcher, self).__init__(**kwargs)
        self.bundles_root = bundles_root
        self.remotes = remotes

        # For debugging: in case you need to do something with the transaction manager
        # used when populating the indexed database
        self._transaction_manager = (transaction_manager or
                transaction.TransactionManager())

    def __call__(self, *args, **kwargs):
        '''
        Calls `fetch` with the given arguments
        '''
        return self.fetch(*args, **kwargs)

    def fetch(self, bundle_id, bundle_version=None, remotes=None, progress_reporter=None,
            triples_progress_reporter=None):
        '''
        Retrieve a bundle by name from a remote and put it in the local bundle cache.

        The first remote that can retrieve the bundle will be tried. Each remote will be
        tried in succession until one downloads the bundle.

        Parameters
        ----------
        bundle_id : str
            The id of the bundle to retrieve
        bundle_version : int, optional
            The version of the bundle to retrieve. If not provided, attempt to fetch the
            latest version available
        remotes : iterable of Remote or str
            A subset of remotes and additional remotes to fetch from. If an entry in the
            iterable is a string, then it will be looked for amongst the remotes passed in
            initially.
        progress_reporter : `tqdm.tqdm <https://tqdm.github.io/>`_-like object, optional
            Receives updates of progress in fetching and installing locally
        triples_progress_reporter : `tqdm.tqdm <https://tqdm.github.io/>`_-like object, optional
            Receives updates of progress for adding triples for an individual graph

        Returns
        -------
        str
            returns the directory where the bundle has been placed

        Raises
        ------
        .exceptions.NoBundleLoader
            Thrown when none of the loaders are able to download the bundle
        .FetchTargetIsNotEmpty
            Thrown when the requested bundle is already in the cache
        '''
        if remotes is not None:
            remotes = list(remotes)
        given_bundle_version = bundle_version
        loaders = self._get_bundle_loaders(bundle_id, given_bundle_version, remotes)
        loaders_list = list(loaders)

        if bundle_version is None:
            bundle_version = self._find_latest_remote_bundle_versions(bundle_id, loaders_list)

        bdir = fmt_bundle_directory(self.bundles_root, bundle_id, bundle_version)
        self._assert_target_is_empty(bdir)

        for loader in loaders_list:
            try:
                loader.base_directory = bdir
                loader(bundle_id, bundle_version)
                with open(p(bdir, BUNDLE_MANIFEST_FILE_NAME)) as mf:
                    manifest_data = json.load(mf)
                    for dd in manifest_data.get('dependencies', ()):
                        try:
                            find_bundle_directory(self.bundles_root, dd['id'], dd.get('version'))
                        except BundleNotFound:
                            self.fetch(dd['id'], dd.get('version'), remotes=remotes)
                dat = self._post_fetch_dest_conf(bdir)
                build_indexed_database(dat['rdf.graph'], bdir, self._transaction_manager,
                        progress_reporter,
                        triples_progress_reporter)
                dat.close()
                return bdir
            except Exception:
                L.warning('Failed to load bundle %s with %s', bundle_id, loader, exc_info=True)
                shutil.rmtree(bdir)
        else:  # no break
            raise NoBundleLoader(bundle_id, given_bundle_version)

    def _post_fetch_dest_conf(self, bundle_directory):
        db_file_location = p(bundle_directory, BUNDLE_INDEXED_DB_NAME)
        store_conf = dict(url=db_file_location,
                transaction_manager=self._transaction_manager)
        res = Data().copy({
            'rdf.source': 'default',
            'rdf.store': 'FileStorageZODB',
            'rdf.store_conf': store_conf
        })
        res.init()
        if not exists(db_file_location):
            raise Exception('Could not create the database file at ' + db_file_location)
        return res

    def _find_latest_remote_bundle_versions(self, bundle_id, loaders_list):
        latest_bundle_version = 0
        for loader in loaders_list:
            try:
                versions = loader.bundle_versions(bundle_id)
            except NotImplementedError:
                L.debug('Loader %s does not support listing available versions of %s',
                        loader, bundle_id)
                continue
            if not versions:
                L.warning('Loader %s does not have any versions of the bundle %s', loader, bundle_id)
                continue
            loader_latest_version = max(versions)
            if loader_latest_version > latest_bundle_version:
                latest_bundle_version = loader_latest_version
        if latest_bundle_version <= 0:
            raise BundleNotFound(bundle_id, 'No versions of the requested bundle found from any remotes')
        return latest_bundle_version

    def _assert_target_is_empty(self, bdir):
        target_empty = True
        try:
            for _ in scandir(bdir):
                target_empty = False
                break
        except FileNotFoundError:
            return
        if not target_empty:
            raise FetchTargetIsNotEmpty(bdir)

    def _get_bundle_loaders(self, bundle_id, bundle_version, remotes):
        try:
            retrieved_remotes = self._get_remotes(remotes)
        except NoRemoteAvailable as e:
            raise NoBundleLoader(bundle_id, bundle_version,
                    'Could not get any remotes to generate loaders from') from e
        else:
            for rem in retrieved_remotes:
                for loader in rem.generate_loaders():
                    if loader.can_load(bundle_id, bundle_version):
                        yield loader

    def __str__(self):
        return f'{self.__class__.__name__}(bundles_root={self.bundles_root}, remotes={self.remotes})'


class Deployer(_RemoteHandlerMixin):
    '''
    Deploys bundles to `Remotes <Remote>`.

    A deployer takes a bundle directory tree or bundle archive and uploads it to a remote.
    `Fetcher` is, functionally, the dual of this class.

    Deployer is responsible for selecting remotes and corresponding uploaders among a set
    of options. `Uploaders <Uploader>` are responsible for actually doing the upload.
    '''

    def __init__(self, remotes=(), **kwargs):
        super(Deployer, self).__init__(**kwargs)
        self.remotes = remotes

    def __call__(self, *args, **kwargs):
        return self.deploy(*args, **kwargs)

    def deploy(self, bundle_path, remotes=None):
        '''
        Deploy a bundle to *all* remotes that are configured to accept uploads

        Parameters
        ----------
        bundle_path : str
            Path to a bundle directory tree or archive
        remotes : iterable of Remote or str
            A subset of remotes to deploy to and additional remotes to deploy to

        Raises
        ------
        .NoAcceptableUploaders
            Thrown when none of the selected uploaders could upload the bundle
        '''
        if not exists(bundle_path):
            raise NotABundlePath(bundle_path, 'the file does not exist')

        manifest_data = self._extract_manifest_data_from_bundle_path(bundle_path)
        validate_manifest(bundle_path, manifest_data)

        uploaded = False
        for uploader in self._get_bundle_uploaders(bundle_path, remotes=remotes):
            uploader(bundle_path)
            uploaded = True

        if not uploaded:
            raise NoAcceptableUploaders(bundle_path)

    def _extract_manifest_data_from_bundle_path(self, bundle_path):
        if isdir(bundle_path):
            return self._get_directory_manifest_data(bundle_path)
        elif isfile(bundle_path):
            return self._get_archive_manifest_data(bundle_path)
        else:
            raise NotABundlePath(bundle_path, 'path does not point to a file or directory')

    def _get_bundle_uploaders(self, bundle_directory, remotes=None):
        for rem in self._get_remotes(remotes):
            for uploader in rem.generate_uploaders():
                if uploader.can_upload(bundle_directory):
                    yield uploader

    def _get_directory_manifest_data(self, bundle_path):
        try:
            with open(p(bundle_path, BUNDLE_MANIFEST_FILE_NAME)) as mf:
                return json.load(mf)
        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT: # FileNotFound
                raise MalformedBundle(bundle_path, 'no bundle manifest found')
            if e.errno == errno.EISDIR: # IsADirectoryError
                raise MalformedBundle(bundle_path, 'manifest is not a regular file')
            raise
        except json.decoder.JSONDecodeError:
            raise MalformedBundle(bundle_path, 'manifest is malformed: expected a'
                    ' JSON file')

    def _get_archive_manifest_data(self, bundle_path):
        with Unarchiver().to_tarfile(bundle_path) as tf:
            try:
                mf0 = tf.extractfile(BUNDLE_MANIFEST_FILE_NAME)
                if mf0 is None:
                    raise MalformedBundle(bundle_path, 'manifest is not a regular file')
                # Would like to pull the
                with mf0 as mf:
                    return json.load(mf)
            except KeyError:
                raise MalformedBundle(bundle_path, 'no bundle manifest found')
            except json.decoder.JSONDecodeError:
                raise MalformedBundle(bundle_path, 'manifest is malformed: expected a'
                        ' JSON file')


class Cache(object):
    '''
    Cache of bundles
    '''

    def __init__(self, bundles_directory):
        '''
        Parameters
        ----------
        bundles_directory : str
            The where bundles are stored
        '''
        self.bundles_directory = bundles_directory

    def list(self):
        '''
        Returns a generator of summary bundle info
        '''
        try:
            bundle_directories = scandir(self.bundles_directory)
        except (OSError, IOError) as e:
            if e.errno == errno.ENOENT:
                return
            raise

        for bundle_directory in bundle_directories:
            if not bundle_directory.is_dir():
                continue

            # Ignore deletes out from under us
            try:
                version_directories = scandir(bundle_directory.path)
            except (OSError, IOError) as e:
                if e.errno == errno.ENOENT:
                    continue
                raise

            def keyfunc(x):
                try:
                    return int(x.name)
                except ValueError:
                    return float('+inf')

            for version_directory in sorted(version_directories, key=keyfunc, reverse=True):
                if not version_directory.is_dir():
                    continue
                try:
                    manifest_fname = p(version_directory.path, BUNDLE_MANIFEST_FILE_NAME)
                    with open(manifest_fname) as mf:
                        try:
                            manifest_data = json.load(mf)
                            bd_id = urlunquote(bundle_directory.name)
                            bd_version = int(version_directory.name)
                            if (bd_id != manifest_data.get('id') or
                                    bd_version != manifest_data.get('version')):
                                L.warning('Bundle manifest at %s does not match bundle'
                                ' directory', manifest_fname)
                                continue
                            yield manifest_data
                        except json.decoder.JSONDecodeError:
                            L.warning("Bundle manifest at %s is malformed",
                                   manifest_fname)
                except (OSError, IOError) as e:
                    if e.errno != errno.ENOENT:
                        raise


def retrieve_remote_by_name(remotes_dir, name, **kwargs):
    for rem in retrieve_remotes(remotes_dir, **kwargs):
        if rem.name == name:
            return rem


def retrieve_remotes(remotes_dir, load_entry_points=True):
    '''
    Retrieve remotes from a project directory or user remotes directory

    Parameters
    ----------
    owmdir : str
        path to the project directory
    load_entry_points : bool, optional
        if `True`, then the entry points for `~.loaders.Loader` and `~.loaders.Uploader`
        implementations that have been added as entry points
    '''
    if not exists(remotes_dir):
        return

    if load_entry_points:
        load_entry_point_loaders()

    for r in listdir(remotes_dir):
        if r.endswith('.remote'):
            fname = p(remotes_dir, r)
            with open(fname) as inp:
                try:
                    rem = Remote.read(inp)
                    rem.file_name = fname
                    yield rem
                except Exception:
                    L.warning('Unable to read remote %s', r, exc_info=True)


class Installer(object):
    '''
    Installs a bundle locally
    '''

    def __init__(self, source_directory, bundles_directory, graph,
                 imports_ctx=None, default_ctx=None, class_registry_ctx=None,
                 installer_id=None, remotes=(), remotes_directory=None):
        '''
        Parameters
        ----------
        source_directory : str
            Directory where files come from. All files for a bundle must be below this
            directory
        bundles_directory : str
            Directory where the bundles files go. Usually this is the bundle cache
            directory
        graph : rdflib.graph.ConjunctiveGraph
            The graph from which we source contexts for this bundle
        default_ctx : str, optional
            The ID of the default context -- the target of a query when not otherwise
            specified.
        imports_ctx : str, optional
            The ID of the imports context this installer should use. Imports relationships
            are selected from this graph according to the included contexts.
        class_registry_ctx : str, optional
            The ID of the class registry context this installer should use. Class registry
            entries are retrieved from this graph.
        installer_id : iterable of Remote or str, optional
            Name of this installer for purposes of mutual exclusion
        remotes : iterable of Remote, optional
            Remotes to be used for retrieving dependencies when needed during
            installation. If not provided, the remotes will be collected from
            `remotes_directory`
        remotes_directory : str, optional
            The directory to load `Remotes <Remote>` from in case a bundle is not in the
            bundle cache. Defaults to `.DEFAULT_REMOTES_DIRECTORY`
        '''
        self.context_hash = hashlib.sha224
        self.file_hash = hashlib.sha224
        self.source_directory = source_directory
        self.bundles_directory = bundles_directory
        self.graph = graph
        self.installer_id = installer_id
        self.imports_ctx = imports_ctx
        self.default_ctx = default_ctx
        self.class_registry_ctx = class_registry_ctx
        self.remotes = list(remotes)
        self.remotes_directory = remotes_directory

        # For debugging: in case you need to do something with the transaction manager
        # used when populating the indexed database
        self._transaction_manager = transaction.TransactionManager()

    def install(self, descriptor, progress_reporter=None):
        '''
        Given a descriptor, install a bundle

        Parameters
        ----------
        descriptor : Descriptor
            The descriptor for the bundle
        progress_reporter : `tqdm.tqdm <https://tqdm.github.io/>`_-like object
            Used for reporting progress during installation. optional

        Returns
        -------
        str
            The directory where the bundle is installed

        Raises
        ------
        .TargetIsNotEmpty
            Thrown when the target directory for installation is not empty.
        '''
        # Create the staging directory in the base directory to reduce the chance of
        # moving across file systems
        try:
            staging_directory = fmt_bundle_directory(self.bundles_directory, descriptor.id,
                    descriptor.version)
            makedirs(staging_directory)
        except OSError:
            pass

        target_empty = True
        for _ in scandir(staging_directory):
            target_empty = False
            break
        if not target_empty:
            raise TargetIsNotEmpty(staging_directory)

        with lock_file(p(staging_directory, '.lock'), unique_key=self.installer_id):
            try:
                self._install(descriptor, staging_directory,
                        progress_reporter=progress_reporter)
                return staging_directory
            except Exception:
                self._cleanup_failed_install(staging_directory)
                raise

    def _cleanup_failed_install(self, staging_directory):
        shutil.rmtree(p(staging_directory, 'graphs'))
        shutil.rmtree(p(staging_directory, 'files'))

    def _install(self, descriptor, staging_directory, progress_reporter=None):
        graphs_directory, files_directory = self._set_up_directories(staging_directory)
        self._write_file_hashes(descriptor, files_directory)
        self._write_context_data(descriptor, graphs_directory)
        self._generate_bundle_class_registry_ctx(descriptor, graphs_directory)
        self._generate_bundle_imports_ctx(descriptor, graphs_directory)
        self._write_manifest(descriptor, staging_directory)
        self._initdb(staging_directory)
        self._build_indexed_database(staging_directory, progress_reporter)

    def _set_up_directories(self, staging_directory):
        graphs_directory = p(staging_directory, 'graphs')
        files_directory = p(staging_directory, 'files')

        try:
            makedirs(graphs_directory)
            makedirs(files_directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        return graphs_directory, files_directory

    def _write_file_hashes(self, descriptor, files_directory):
        with open(p(files_directory, 'hashes'), 'wb') as hash_out:
            for fname in _select_files(descriptor, self.source_directory):
                hsh = self.file_hash()
                source_fname = p(self.source_directory, fname)
                hash_file(hsh, source_fname)
                self._write_hash_line(hash_out, fname.encode('UTF-8'), hsh)
                shutil.copy2(source_fname, p(files_directory, fname))

    def _write_context_data(self, descriptor, graphs_directory):
        contexts = _select_contexts(descriptor, self.graph)
        imports_ctxg = None
        if self.imports_ctx:
            imports_ctxg = self.graph.get_context(self.imports_ctx)

        included_context_ids = set()

        for ctxid in self._write_graphs(graphs_directory, *contexts):
            included_context_ids.add(ctxid)

        for c in descriptor.empties:
            included_context_ids.add(URIRef(c))

        # Compute imported contexts
        imported_contexts = set()
        for ctxid in included_context_ids:
            if imports_ctxg is not None:
                for t in imports_ctxg.triples((ctxid, CONTEXT_IMPORTS, None)):
                    imported_contexts.add(t[2])
        uncovered_contexts = imported_contexts - included_context_ids
        if self.class_registry_ctx:
            uncovered_contexts.discard(URIRef(self.class_registry_ctx))
        uncovered_contexts = self._cover_with_dependencies(uncovered_contexts, descriptor)
        if uncovered_contexts:
            raise UncoveredImports(uncovered_contexts)

    def _write_manifest(self, descriptor, staging_directory):
        manifest_data = {}
        if self.default_ctx:
            manifest_data[DEFAULT_CONTEXT_KEY] = self.default_ctx

        if self.imports_ctx:
            # If an imports context was specified, then we'll need to generate an
            # imports context with the appropriate imports. We don't use the source
            # imports context ID for the bundle's imports context because the bundle
            # imports that we actually need are a subset of the total set of imports
            manifest_data[IMPORTS_CONTEXT_KEY] = fmt_bundle_imports_ctx_id(descriptor.id,
                    descriptor.version)

        if self.class_registry_ctx:
            manifest_data[CLASS_REGISTRY_CONTEXT_KEY] = fmt_bundle_class_registry_ctx_id(descriptor.id,
                    descriptor.version)

        manifest_data['id'] = descriptor.id
        manifest_data['version'] = descriptor.version
        manifest_data['manifest_version'] = BUNDLE_MANIFEST_VERSION
        mf_deps = []
        for dd in descriptor.dependencies:
            bnd = self._dd_to_bundle(dd)
            # Fetch the dependency if necessary and get the version of the latest from the
            # bundle manifest. Usually, the bundle will already be on the system since it
            # *should have* been used for testing.
            #
            # (It's probably possible to do something like just grabbing the bundle
            # manifest data in the case there is not a local copy of the bundle, but that
            # should be unusual enough that it's probably not justified considering the
            # overhead of having an alternative to fetching that bundle loaders might be
            # expected to support.)
            dd_version = bnd.manifest_data['version']
            mf_deps.append({'version': dd_version,
                'id': dd.id,
                'excludes': dd.excludes})
        manifest_data['dependencies'] = mf_deps
        self.manifest_data = manifest_data
        with open(p(staging_directory, BUNDLE_MANIFEST_FILE_NAME), 'w') as mf:
            json.dump(manifest_data, mf, separators=(',', ':'))

    def _generate_bundle_imports_ctx(self, descriptor, graphs_directory):
        if not self.imports_ctx:
            return
        imports_ctxg = self.graph.get_context(self.imports_ctx)
        # select all of the imports for all of the contexts in the bundle and serialize
        contexts = []
        idx_fname = p(graphs_directory, 'index')
        with open(idx_fname) as index_file:
            for l in index_file:
                ctx, _ = l.strip().split('\x00')
                contexts.append(URIRef(ctx))
        for c in descriptor.empties:
            contexts.append(URIRef(c))
        if self.class_registry_ctx:
            contexts.append(URIRef(self.class_registry_ctx))
        ctxgraph = imports_ctxg.triples_choices((contexts, CONTEXT_IMPORTS, None))
        if self.class_registry_ctx:
            cr_ctxid = URIRef(fmt_bundle_class_registry_ctx_id(descriptor.id, descriptor.version))
            old_ctxgraph = ctxgraph

            def replace_cr_ctxid():
                src_cr_ctxid = URIRef(self.class_registry_ctx)
                for t in old_ctxgraph:
                    if t[0] == src_cr_ctxid:
                        yield (cr_ctxid, t[1], t[2])
                    elif t[2] == src_cr_ctxid:
                        yield (t[0], t[1], cr_ctxid)
                    else:
                        yield t
            ctxgraph = replace_cr_ctxid()

        ctxid = fmt_bundle_imports_ctx_id(descriptor.id, descriptor.version)
        self._write_graph(graphs_directory, ctxid, ctxgraph)

    def _generate_bundle_class_registry_ctx(self, descriptor, graphs_directory):
        if not self.class_registry_ctx:
            return
        ctx_id = fmt_bundle_class_registry_ctx_id(descriptor.id, descriptor.version)
        class_registry_ctxg = self.graph.get_context(self.class_registry_ctx)

        self._write_graph(graphs_directory, ctx_id, class_registry_ctxg)

    def _write_graph(self, graphs_directory, ctxid, ctxgraph):
        for _ in self._write_graphs(graphs_directory, (ctxid, ctxgraph)):
            pass

    def _write_graphs(self, graphs_directory, *graphs_sequence):
        with open(p(graphs_directory, 'hashes'), 'ab') as hash_out,\
                open(p(graphs_directory, 'index'), 'ab') as index_out:
            for ctxid, ctxgraph in graphs_sequence:
                ctxidb = ctxid.encode('UTF-8')
                gbname, hsh = self._write_graph_to_file(ctxgraph, graphs_directory)
                self._write_hash_line(hash_out, ctxidb, hsh)
                self._write_index_line(index_out, ctxidb, gbname)
                yield ctxid
            hash_out.flush()
            index_out.flush()

    def _write_graph_to_file(self, ctxgraph, graphs_directory):
        hsh = self.context_hash()
        temp_fname = p(graphs_directory, 'graph.tmp')
        write_canonical_to_file(ctxgraph, temp_fname)
        hash_file(hsh, temp_fname)
        gbname = hsh.hexdigest() + '.nt'
        ctx_file_name = p(graphs_directory, gbname)
        rename(temp_fname, ctx_file_name)
        return gbname, hsh

    def _write_hash_line(self, hash_out, key, hsh):
        hash_out.write(key + b'\x00' + pack('B', hsh.digest_size) + hsh.digest() + b'\n')

    def _write_index_line(self, index_out, ctxidb, gbname):
        index_out.write(ctxidb + b'\x00' + gbname.encode('UTF-8') + b'\n')

    def _initdb(self, staging_directory):
        db_file_location = p(staging_directory, BUNDLE_INDEXED_DB_NAME)
        self.conf = Data().copy({
            'rdf.source': 'default',
            'rdf.store': 'FileStorageZODB',
            'rdf.store_conf': dict(transaction_manager=self._transaction_manager,
                url=db_file_location)
        })
        # Create the database file and initialize some needed data structures
        self.conf.init()
        if not exists(db_file_location):
            raise Exception('Could not create the database file at ' + db_file_location)

    def _build_indexed_database(self, staging_directory, progress=None):
        try:
            dest = self.conf['rdf.graph']
            build_indexed_database(dest, staging_directory, self._transaction_manager,
                    progress, None)
        finally:
            self.conf.close()

    def _dd_to_bundle(self, dependency_descriptor):
        return Bundle(dependency_descriptor.id,
                version=dependency_descriptor.version,
                bundles_directory=self.bundles_directory,
                remotes=self.remotes,
                remotes_directory=self.remotes_directory)

    def _cover_with_dependencies(self, uncovered_contexts, descriptor):
        # XXX: Will also need to check for the contexts having a given ID being consistent
        # with each other across dependencies
        dependencies = descriptor.dependencies
        for d in dependencies:
            bnd = self._dd_to_bundle(d)
            for c in bnd.contexts:
                uncovered_contexts.discard(URIRef(c))
                if not uncovered_contexts:
                    break
        return uncovered_contexts


def fmt_bundle_imports_ctx_id(id, version):
    return fmt_bundle_ctx_id('generated_imports_ctx', id, version)


def fmt_bundle_class_registry_ctx_id(id, version):
    return fmt_bundle_ctx_id('generated_class_registry_ctx', id, version)


def fmt_bundle_class_registry_ctx_list_id(id, version):
    return fmt_bundle_ctx_id('generated_class_registry_ctx_list', id, version)


def fmt_bundle_ctx_id(kind, id, version):
    return f'http://data.openworm.org/bundle/{kind}?bundle_id={urlquote(id)}&bundle_version={version}'


class FilesDescriptor(object):
    '''
    Descriptor for files
    '''
    def __init__(self):
        self.patterns = set()
        self.includes = set()

    @classmethod
    def make(cls, obj):
        if not obj:
            return None
        res = cls()
        res.patterns = set(obj.get('patterns', ()))
        res.includes = set(obj.get('includes', ()))
        return res

    def to_dict(self):
        res = {}
        if res.patterns:
            res['patterns'] = list(res.patterns)

        if res.includes:
            res['includes'] = list(res.includes)

        return res


def make_pattern(s):
    if s.startswith('rgx:'):
        return RegexURIPattern(s[4:])
    else:
        return GlobURIPattern(s)


def make_include_func(s):
    if isinstance(s, str):
        return URIIncludeFunc(s)
    elif isinstance(s, dict):
        uri = None
        for k in s.keys():
            if uri is not None:
                raise ValueError('Context "includes" entry must have one key--the URI of'
                        f' the context to include. Extra key is "{k}"')
            uri = k

        return URIIncludeFunc(uri)
    else:
        raise ValueError('Context "includes" entry must be a str or a dict')


class URIIncludeFunc(object):

    def __init__(self, include):
        self.include = URIRef(include.strip())

    def __hash__(self):
        return hash(self.include)

    def __call__(self, uri):
        return URIRef(uri.strip()) == self.include

    def __str__(self):
        return '{}({})'.format(FCN(type(self)), repr(self.include))

    __repr__ = __str__


class URIPattern(object):
    def __init__(self, pattern):
        self._pattern = pattern

    def __hash__(self):
        return hash(self._pattern)

    def __call__(self, uri):
        return False

    def __str__(self):
        return '{}({})'.format(FCN(type(self)), self._pattern)


class RegexURIPattern(URIPattern):
    def __init__(self, pattern):
        super(RegexURIPattern, self).__init__(re.compile(pattern))

    def __call__(self, uri):
        # Cast the pattern match result to a boolean
        return not not self._pattern.match(str(uri))


class GlobURIPattern(RegexURIPattern):
    def __init__(self, pattern):
        replacements = [
            ['*', '.*'],
            ['?', '.?'],
            ['[!', '[^']
        ]

        for a, b in replacements:
            pattern = pattern.replace(a, b)
        super(GlobURIPattern, self).__init__(re.compile(pattern))


def _select_files(descriptor, directory):
    fdescr = descriptor.files
    if not fdescr:
        return
    for f in fdescr.includes:
        if not exists(p(directory, f)):
            raise Exception('Included file in bundle does not exist', f)
        yield f

    for f in fdescr.patterns:
        for match in match_files(directory, p(directory, f)):
            yield relpath(match, directory)


def _select_contexts(descriptor, graph):
    for context in graph.contexts():
        ctx = context.identifier
        for inc in descriptor.includes:
            if inc(ctx):
                yield ctx, context
                break

        for pat in descriptor.patterns:
            if pat(ctx):
                yield ctx, context
                break


def build_indexed_database(dest, bundle_directory, transaction_manager,
        progress=None, trip_prog=None):
    '''
    Build the indexed database from a bundle directory
    '''
    idx_fname = p(bundle_directory, 'graphs', 'index')
    # This code was copied from OWM._load_all_graphs, but we don't have a specific
    # reason for projects and bundles to have the same format, so keeping the logic
    # separate
    triples_read = 0
    with open(idx_fname) as index_file:
        cnt = 0
        for l in index_file:
            cnt += 1
        index_file.seek(0)
        if progress is not None:
            progress.total = cnt
        with transaction_manager:
            bag = BatchAddGraph(dest, batchsize=10000)
            for l in index_file:
                ctx, fname = l.strip().split('\x00')
                parser = plugin.get('nt', Parser)()
                graph_fname = p(bundle_directory, 'graphs', fname)
                with open(graph_fname, 'rb') as f, bag.get_context(ctx) as g:
                    parser.parse(create_input_source(f), g)

                if progress is not None:
                    progress.update(1)
                if trip_prog is not None:
                    trip_prog.update(bag.count - triples_read)
                triples_read = g.count
            if progress is not None:
                progress.write('Finalizing writes to database...')
    if progress is not None:
        progress.write('Loaded {:,} triples'.format(triples_read))


class BundleTransactionManager(transaction.TransactionManager):
    '''
    Marker class useful in debugging to identify which txn manager we're using
    '''



from __future__ import print_function
from __future__ import absolute_import

from collections import OrderedDict, defaultdict
import logging

from rdflib.term import URIRef
import six

from . import BASE_CONTEXT
from .utils import FCN
from .context import Context
from .dataobject import (DataObject, ObjectProperty, DatatypeProperty, UnionProperty, This,
                         CPThunk)
from .data_trans.common_data import DS_NS, DS_DATA_NS

L = logging.getLogger(__name__)

INFO_PROP_PREFIX = '_info_prop_'


class FormatUtil(object):
    @staticmethod
    def collect_values(attr, stored):
        if stored:
            attr_vals = list()
            for x in attr.get():
                if x not in attr_vals:
                    attr_vals.append(x)
        else:
            attr_vals = attr.defined_values
        return attr_vals


class Informational(object):
    '''
    Defines a property on a `.DataSource`

    Attributes
    ----------
    name : str
        The name for the property
    description : str
        A description of the property
    also : tuple of Informational or list of Informational
        Other properties which, if set, set the value for this property. If multiple such
        "also" properties are set when the owning `DataSource` instance is defined, then
        a `DuplicateAlsoException` will be raised.
    default_override : object
        An override for the default value, typically set by setting the value in a
        `.DataSource` class dictionary. Importantly, this overrides an "also" value which
        would normally take precedence.
    default_value : object
        Default value if no other value is set
    multiple : boolean
        If `True`, then the property can take on multiple values for the same subject
    cls : type
        The `~owmeta_core.dataobject_property.Property` corresponding to this property
    '''

    def __init__(self, name=None, display_name=None, description=None,
                 default_value=None, property_type='DatatypeProperty',
                 multiple=True, property_name=None, also=(), subproperty_of=None,
                 **property_args):
        '''
        Parameters
        ----------
        name : str, optional
            Name for the property. If not provided here, then the name generally gets set
            to the name to which this object is assigned
        display_name : str, optional
            Display name for the property. If not provided here, then the `name` will be
            used for the display name
        description : str, optional
            A description of the property
        default_value : object, optional
            Value to use
        property_type : 'DatatypeProperty', 'ObjectProperty', or 'UnionProperty'
            The type of `~owmeta_core.dataobject_property.Property` to create from this object.
            Default is 'DatatypeProperty'
        multiple : boolean, optional
            Whether this property can have multiple values for the same object. Default is
            `True`
        property_name : str, optional
            The name of the property to use for attributes. `name` will be used if a value
            is not provided here
        also : Informational, tuple of Informational, or list of Informational; optional
            Other properties which, if set, will give their value to this property as well
        subproperty_of : Informational
            Declares that given Informational's corresponding Property is a subproperty of
        **property_args
            Additional arguments which will be passed into the class dictionary when the
            `~owmeta_core.dataobject_property.Property` corresponding to this object is created.
        '''
        self.name = name
        self._property_name = property_name
        self._display_name = display_name
        self.default_value = default_value
        self.description = description
        self.property_type = property_type
        self.multiple = multiple
        if also and not isinstance(also, (list, tuple)):
            also = (also,)
        self.also = also
        self.property_args = property_args

        self.default_override = None

        self.cls = None
        self.subproperty_of = subproperty_of
        self._docstr = None

    @property
    def __doc__(self):
        return (self._docstr or (f'"{self.display_name}", a :class:`~owmeta_core.dataobject.{self.property_type}`' +
            (f': {self.description}' if self.description else '') +
            (f'\n\nDefault value: {self.default_value!r}' if self.default_value is not None else '')))

    @__doc__.setter
    def __doc__(self, docstring):
        self._docstr = docstring

    def __get__(self, obj, owner):
        if obj is None:
            return self
        else:
            return getattr(obj, INFO_PROP_PREFIX + self.name)

    @property
    def display_name(self):
        '''
        The display name for the property.
        '''
        return self._display_name if self._display_name is not None else self.name

    @display_name.setter
    def display_name(self, val):
        self._display_name = val

    @property
    def property_name(self):
        '''
        The name of the property to use for attributes
        '''
        return self._property_name if self._property_name is not None else self.name

    @property_name.setter
    def property_name(self, v):
        self._property_name = v

    def copy(self):
        '''
        Copy to a new `Informational`
        '''
        res = type(self)()
        for x in vars(self):
            setattr(res, x, getattr(self, x))
        return res

    def __repr__(self):
        return ("Informational(name='{}',"
                " display_name={},"
                " default_value={},"
                " description={})").format(self.name,
                                          repr(self.display_name),
                                          repr(self.default_value),
                                          repr(self.description))

    # NOTE: This guy has to come last to avoid conflict with the decorator
    @property
    def property(self):
        return getattr(self.cls, INFO_PROP_PREFIX + self.name).property


class DuplicateAlsoException(Exception):
    pass


class DataSourceType(type(DataObject)):
    """A type for DataSources

    Sets up the graph with things needed for MappedClasses
    """

    def __init__(self, name, bases, dct):
        self.__info_fields = []
        others = []
        newdct = dict()
        keys = dct.keys()
        phase = 0
        while keys and phase < 2:
            unhandled_keys = list()
            for z in keys:
                meta = dct[z]
                if isinstance(meta, Informational):
                    if meta.cls is not None:
                        L.debug("Already created a Property from %s for %s. Not creating another for %s",
                                meta, meta.cls, self)
                        prop_name = INFO_PROP_PREFIX + meta.name
                        meta_owner_property_property = None
                        meta_owner_property = None
                        try:
                            meta_owner_property_property = getattr(meta.cls, prop_name)
                        except AttributeError:
                            if phase == 0:
                                try:
                                    meta_owner_property = newdct[prop_name]
                                except KeyError:
                                    L.debug('Unable to handle Informational %s on %s -- probably a reference to an'
                                            ' Informational defined on this same DataSource. Will re-process.',
                                            meta, self)
                                    unhandled_keys.append(z)
                                    continue
                            else:
                                raise

                        if meta_owner_property:
                            newdct[INFO_PROP_PREFIX + z] = meta_owner_property
                        else:
                            newdct[INFO_PROP_PREFIX + z] = CPThunk(meta_owner_property_property.property)

                        meta_copy = meta.copy()
                        meta_copy.cls = self
                        meta_copy.name = z
                        self.__info_fields.append(meta_copy)
                        setattr(self, z, meta_copy)
                    else:
                        meta.cls = self
                        meta.name = z
                        self.__info_fields.append(meta)

                        # Make the owmeta_core property
                        #
                        # We set the name for the property to the inf.name since that's how we
                        # access the info on this object, but the inf.property_name is used for
                        # the linkName so that the property's URI is generated based on that name.
                        # This allows to set an attribute named inf.property_name on self while
                        # still having access to the property through inf.name.
                        ptype = None
                        if meta.property_type == 'DatatypeProperty':
                            ptype = DatatypeProperty
                        elif meta.property_type == 'ObjectProperty':
                            ptype = ObjectProperty
                        elif meta.property_type == 'UnionProperty':
                            ptype = UnionProperty
                        else:
                            raise ValueError(f'Unrecognized property type {meta.property_type}')

                        property_args = dict(**meta.property_args)
                        superproperty = meta.subproperty_of
                        if isinstance(superproperty, Informational):
                            superproperty_property = newdct.get(INFO_PROP_PREFIX + superproperty.name)
                            if not superproperty_property:
                                try:
                                    superproperty_property = superproperty.property
                                except AttributeError:
                                    raise ValueError(f'{superproperty} is missing a Property definition')
                            property_args['subproperty_of'] = superproperty_property
                        elif isinstance(superproperty, (list, tuple)):
                            sps = []
                            for sp in superproperty:
                                superproperty_property = newdct.get(INFO_PROP_PREFIX + sp.name)
                                if not superproperty_property:
                                    try:
                                        superproperty_property = sp.property
                                    except AttributeError:
                                        raise ValueError(f'{sp} is missing a Property definition')
                                sps.append(superproperty_property)
                            property_args['subproperty_of'] = sps
                        elif meta.subproperty_of:
                            property_args['subproperty_of'] = meta.subproperty_of

                        newdct[INFO_PROP_PREFIX + meta.name] = ptype(
                                linkName=meta.property_name,
                                multiple=meta.multiple,
                                **property_args)
                else:
                    others.append((z, dct[z]))
            keys = unhandled_keys
            phase += 1

        for x in bases:
            if isinstance(x, DataSourceType):
                self.__info_fields += [inf.copy() for inf in x.__info_fields]

        for k, v in others:
            for i in range(len(self.__info_fields)):
                if self.__info_fields[i].name == k:
                    # This is for setting default values from a super-class. We copy the
                    # Informational because the default is baked-in to the Informational
                    # instance, and we want it to apply only to the sub-class
                    self.__info_fields[i].default_override = v
                    setattr(self, k, self.__info_fields[i])
                    break
            else: # no 'break'
                newdct[k] = v
        if not getattr(self, '__doc__', None):
            self.__doc__ = self._docstr()
        super(DataSourceType, self).__init__(name, bases, newdct)

    def _docstr(self):
        s = ''
        for inf in self.__info_fields:
            s += '{} : :class:`~owmeta_core.dataobject.{}`'.format(inf.display_name, inf.property_type) + \
                    ('\n    Attribute: `{}`'.format(inf.name if inf.property_name is None else inf.property_name)) + \
                    (('\n\n    ' + inf.description) if inf.description else '') + \
                    ('\n\n    Default value: {}'.format(inf.default_value) if inf.default_value is not None else '') + \
                    '\n\n'
        return s

    @property
    def info_fields(self):
        return self.__info_fields


class Transformation(DataObject):
    """
    Record of the how a `DataSource` was produced and the sources of the transformation
    that produced it.  Unlike the 'source' field attached to DataSources, the Translation
    may distinguish different kinds of input source to a transformation.
    """

    class_context = BASE_CONTEXT

    transformer = ObjectProperty()

    def defined_augment(self):
        return self.transformer.has_defined_value() and self.transformer.onedef().defined

    def identifier_augment(self):
        return self.make_identifier(self.transformer.onedef().identifier.n3())


class Translation(Transformation):
    '''
    A transformation where, notionally, the general character of the input is preserved.

    In contrast to just a transformation, a translation wouldn't just pick out, say, one
    record within an input source containing several, but would have an output source with o
    '''
    class_context = BASE_CONTEXT

    translator = ObjectProperty(subproperty_of=Transformation.transformer)


class DataSource(six.with_metaclass(DataSourceType, DataObject)):
    '''
    A source for data that can get translated into owmeta_core objects.

    The value for any field can be passed to `~DataSource.__init__` by name. Additionally,
    if the sub-class definition of a DataSource assigns a value for that field like::

        class A(DataSource):
            some_field = 3

    that value will be used over the default value for the field, but not over
    any value provided to `~DataSource.__init__`.
    '''

    class_context = BASE_CONTEXT

    source = Informational(display_name='Input source',
                           description='The data source that was translated into this one',
                           property_type='ObjectProperty',
                           value_type=This)

    transformation = Informational(display_name='Transformation',
                                   description='Information about the transformation process that created this object',
                                   property_type='ObjectProperty',
                                   value_type=Transformation,
                                   cascade_retract=True)

    translation = Informational(display_name='Translation',
                                description='Information about the translation process that created this object',
                                property_type='ObjectProperty',
                                subproperty_of=transformation,
                                value_type=Translation,
                                cascade_retract=True)

    description = Informational(display_name='Description',
                                description='Free-text describing the data source')

    base_namespace = DS_NS
    base_data_namespace = DS_DATA_NS

    def __init__(self, **kwargs):
        # There's a similar behavior in vanilla DataObject, but that doesn't have default
        # defaults and default-overrides. We don't pass the arguments up to DataObject so
        # the `properties_are_init_args` handling isn't used (whether
        # `properties_are_init_args` is True or False we get bad or incomplete behavior
        # when the property arguments are passed up)
        self.info_fields = OrderedDict((i.name, i) for i in self.__class__.info_fields)
        parent_kwargs = dict()
        new_kwargs = dict()
        for k, v in kwargs.items():
            if k not in self.info_fields:
                parent_kwargs[k] = v
            else:
                new_kwargs[k] = v
        super(DataSource, self).__init__(**parent_kwargs)
        vals = defaultdict(dict)
        for n, inf in self.info_fields.items():
            v = new_kwargs.get(n, None)
            if v is not None:
                vals[n]['i'] = v
            else:
                v = inf.default_value

            if inf.default_override is not None:
                vals[n]['e'] = inf.default_override

            vals[n]['d'] = inf.default_value

            for also in inf.also:
                if v is not None and vals[also.name].setdefault('a', v) != v:
                    raise DuplicateAlsoException('Only one also is allowed')

        for n, vl in vals.items():
            inf = self.info_fields[n]
            v = vl.get('i', vl.get('e', vl.get('a', vl['d'])))
            ctxd_prop = getattr(self, INFO_PROP_PREFIX + inf.name)
            if v is not None:
                ctxd_prop(v)

    def after_transform(self):
        '''
        Called after `Transformer.transform`.

        This method should handle any of the things that should happen for an output data
        source after `Transformer.transform` (or `Translator.translate`). This can include
        things like flushing output to files, closing file handles, and writing triples in
        a Context.

        NOTE: Be sure to call this method via super() in sub-classes
        '''

    def defined_augment(self):
        return self.transformation.has_defined_value() or self.translation.has_defined_value()

    def identifier_augment(self):
        '''
        It doesn't make much sense to have translation and transformation set, so we just
        take the first of them
        '''
        return (self.make_identifier(self.transformation.defined_values[0].identifier.n3())
                or self.make_identifier(self.translation.defined_values[0].identifier.n3()))

    def __str__(self):
        try:
            sio = six.StringIO()
            print(self.__class__.__name__, end='', file=sio)
            if self.defined:
                ident = self.identifier
                if self.namespace_manager:
                    ident = self.namespace_manager.normalizeUri(ident)
                print(f'({ident}', end='', file=sio)

            for info in self.info_fields.values():
                attr = getattr(self, info.name)
                attr_vals = FormatUtil.collect_values(attr, False)
                if attr_vals:
                    print(f', {info.name}=', end='', file=sio)
                    vals = []
                    for val in sorted(attr_vals):
                        if isinstance(val, (DataSource, GenericTranslation)):
                            valstr = str(val)
                        elif isinstance(val, URIRef):
                            valstr = val.n3()
                        elif isinstance(val, six.string_types):
                            valstr = repr(val)
                        else:
                            valstr = str(val)
                        vals.append(valstr)
                    print('[', end='', file=sio)
                    print(', '.join(vals), end='', file=sio)
                    print(']', end='', file=sio)
            print(')', end='', file=sio)
            return sio.getvalue()
        except AttributeError:
            res = super(DataSource, self).__str__()
            L.error('Failed while creating formatting string representation for %s', res,
                    exc_info=True)
            return res

    def format_str(self, stored):
        try:
            sio = six.StringIO()
            print(self.__class__.__name__, end='', file=sio)
            if self.defined:
                ident = self.identifier
                if self.namespace_manager:
                    ident = self.namespace_manager.normalizeUri(ident)
                print(f'({ident})', file=sio)
            else:
                print(file=sio)
            for info in self.info_fields.values():
                attr = getattr(self, info.name)
                attr_vals = FormatUtil.collect_values(attr, stored)
                if attr_vals:
                    print('    ' + info.display_name, end=': ', file=sio)
                    for val in sorted(attr_vals):
                        val_line_sep = '\n      ' + ' ' * len(info.display_name)
                        if isinstance(val, (DataSource, GenericTranslation)):
                            valstr = val.format_str(stored)
                        elif isinstance(val, URIRef):
                            valstr = val.n3()
                        elif isinstance(val, six.string_types):
                            valstr = repr(val)
                        else:
                            valstr = str(val)
                        print(val_line_sep.join(valstr.split('\n')), end=' ', file=sio)
                    print(file=sio)
            return sio.getvalue()
        except AttributeError:
            res = super(DataSource, self).__str__()
            L.error('Failed while creating formatting string representation for %s', res,
                    exc_info=True)
            return res


class OneOrMore(object):
    """
    Wrapper for :class:`DataTransformer` input :class:`DataSource` types indicating that
    one or more of the wrapped type must be provided to the translator
    """
    def __init__(self, source_type):
        self.source_type = source_type

    def __repr__(self):
        return f"{FCN(type(self))}({self.source_type!r})"


class GenericTranslation(Translation):
    """
    A generic translation that just has sources in any order
    """

    class_context = BASE_CONTEXT

    source = ObjectProperty(multiple=True, value_rdf_type=DataSource.rdf_type)

    def defined_augment(self):
        return super(GenericTranslation, self).defined_augment() and \
                self.source.has_defined_value()

    def identifier_augment(self):
        data = super(GenericTranslation, self).identifier_augment().n3() + \
                "".join(sorted(x.identifier.n3() for x in self.source.defined_values))
        return self.make_identifier(data)

    def __str__(self):
        sio = six.StringIO()
        print(f'{self.__class__.__name__}({self.idl}, ', end='', file=sio)
        sources_field_name = 'sources='
        print(sources_field_name, end='', file=sio)

        attr = self.source
        attr_vals = FormatUtil.collect_values(attr, False)

        if attr_vals:
            print(', '.join(str(x) for x in sorted(attr_vals)), end='', file=sio)

        translator = self.translator.onedef()
        if translator is not None:
            print(f'translator={translator}', end='', file=sio)
        return sio.getvalue()

    def format_str(self, stored):
        sio = six.StringIO()
        print(f'{self.__class__.__name__}({self.idl})', file=sio)
        sources_field_name = 'Sources: '
        print(sources_field_name, end='', file=sio)

        attr = self.source
        attr_vals = FormatUtil.collect_values(attr, stored)

        if attr_vals:
            val_line_sep = '\n' + len(sources_field_name) * ' '
            print(val_line_sep.join(val_line_sep.join(val.format_str(stored).split('\n'))
                                    for val in sorted(attr_vals)), file=sio)

        if stored:
            translator = self.translator.one()
        else:
            translator = self.translator.onedef()
        if translator is not None:
            field = "Translator: "
            s = ('\n' + len(field) * ' ').join(str(translator).split('\n'))
            print(field + s, file=sio)
        return sio.getvalue()


class DataObjectContextDataSource(DataSource):

    class_context = BASE_CONTEXT

    def __init__(self, context, **kwargs):
        super(DataObjectContextDataSource, self).__init__(**kwargs)
        if context is not None:
            self.context = context
        else:
            self.context = Context()


def format_types(typ):
    if isinstance(typ, OneOrMore):
        return ':class:`{}` (:class:`~{}`)'.format(FCN(OneOrMore), FCN(typ.source_type))
    elif isinstance(typ, type):
        return ':class:`~{}`'.format(FCN(typ))
    else:
        return ', '.join(':class:`~{}`'.format(FCN(x)) for x in typ)


class DataTransformerType(type(DataObject)):
    def __init__(self, name, bases, dct):
        super(DataTransformerType, self).__init__(name, bases, dct)

        if not getattr(self, '__doc__', None):
            self.__doc__ = f'''Input type(s): {format_types(self.input_type)}\n
                               Output type(s): {format_types(self.output_type)}\n'''


class DataTransformer(six.with_metaclass(DataTransformerType, DataObject)):
    '''
    Transforms one or more `DataSources <DataSource>` to one or more other `DataSources
    <DataSource>`

    Attributes
    ----------
    input_type : `type` or `tuple` of `type`
        Types of input to this transformer. Types should be sub-classes of `DataSource`
    output_type : `type` or `tuple` of `type`
        Types of output from this transformer. Types should be sub-classes of `DataSource`
    transformation_type : type
        Type of the `Transformation` record produced as a side-effect of transforming with
        this transformer
    output_key : str
        The "key" for outputs from this transformer (see `IdentifierMixin`). Normally only
        defined during execution of __call__
    output_identifier : str
        The identifier for outputs from this transformer. Normally only defined during
        execution of __call__
    '''

    class_context = BASE_CONTEXT

    input_type = DataSource
    output_type = DataSource
    transformation_type = Transformation

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_key = None
        self.output_identifier = None

    def __call__(self, *args, output_key=None,
            output_identifier=None, **kwargs):
        self.output_key = output_key
        self.output_identifier = output_identifier
        try:
            res = self.transform(*args, **kwargs)
            res.after_transform()
            res.context.save_context()
            self.after_transform()
            return res
        finally:
            self.output_key = None
            self.output_identifier = None

    def __str__(self):
        s = '''Input type(s): {}
               Output type(s): {}'''.format(self.input_type,
                                            self.output_type)
        return f'{FCN(type(self))}({self.idl})' + ': \n    ' + ('\n    '.join(x.strip() for x in s.split('\n')))

    def defined_augment(self):
        return True

    def identifier_augment(self):
        return self.make_identifier(type(self).rdf_type)

    def transform(self, *args, **kwargs):
        '''
        Notionally, this method takes a data source, which is transformed into some other
        data source. There doesn't necessarily need to be an input data source.

        Parameters
        ----------
        *args
            Input data sources
        **kwargs
            Named input data sources

        Returns
        -------
        the output data source
        '''
        raise NotImplementedError

    def make_transformation(self, sources=()):
        '''
        It's intended that implementations of `DataTransformer` will override this
        method to make custom `Transformations <Transformation>` according with how different
        arguments to `transform` are (or are not) distinguished.

        The actual properties of a `Transformation` subclass must be assigned within the
        `transform` method
        '''
        return self.transformation_type.contextualize(self.context)(transformer=self)

    def after_transform(self):
        '''
        Called after `transform` runs in `__call__` and after the result
        `DataSource.after_transform` is called.
        '''

    def make_new_output(self, sources, *args, **kwargs):
        '''
        Make a new output `DataSource`. Typically called within `transform`.
        '''
        trans = self.make_transformation(sources)

        if self.output_key:
            kwargs['key'] = self.output_key

        if self.output_identifier:
            kwargs['ident'] = self.output_identifier

        res = self.output_type.contextualize(self.context)(*args, transformation=trans,
                                                           conf=self.conf,
                                                           **kwargs)
        for s in sources:
            res.source(s)

        return res

    def transform_with(self, translator_type, *sources, output_key=None,
            output_identifier=None,
            **named_sources):
        '''
        Transform with the given `DataTransformer` and sources.

        This should be used in a `transform` implementation to compose multiple
        transformations. An instance of the transformer will be created and contextualized
        with the *this* transformer's context unless the given transformer already has a
        context.
        '''
        if translator_type.context is None:
            translator_type = translator_type.contextualize(self.context)
        return transform(
                translator_type(),
                output_key=output_key,
                output_identifier=output_identifier,
                data_sources=sources,
                named_data_sources=named_sources)


class BaseDataTranslator(DataTransformer):

    class_context = BASE_CONTEXT

    def translate(self, *args, **kwargs):
        '''
        Notionally, this method takes one or more data sources, and translates them into
        some other data source that captures essentially the same information, but,
        possibly, in a different format. Additional sources can be passed in as well for
        auxiliary information which are not "translated" in their entirety into the output
        data source. Such auxiliarry data sources should be distinguished from the primary
        ones in the translation

        Parameters
        ----------
        *args
            Input data sources
        **kwargs
            Named input data sources

        Returns
        -------
        the output data source
        '''
        raise NotImplementedError

    def transform(self, *args, **kwargs):
        '''
        Just calls `translate` and returns its result.
        '''
        return self.translate(*args, **kwargs)

    def make_translation(self, sources=()):
        '''
        It's intended that implementations of `BaseDataTranslator` will override this
        method to make custom `Translations <Translation>` according with how different
        arguments to `translate` are (or are not) distinguished.

        The actual properties of a `Translation` subclass must be assigned within the
        `translate` method

        Parameters
        ----------
        sources : tuple
            The sources that go into the translation. Sub-classes may choose to pass these
            to their superclass' make_translation method or not.

        Returns
        -------
        a description of the translation

        '''
        return self.translation_type.contextualize(self.context)(transformer=self)

    def make_transformation(self, sources=()):
        '''
        Just calls `make_translation` and returns its result.
        '''
        return self.make_translation(sources)


class DataTranslator(BaseDataTranslator):
    """
    A specialization with the :class:`GenericTranslation` translation type that adds
    sources for the translation automatically when a new output is made
    """

    class_context = BASE_CONTEXT

    translation_type = GenericTranslation

    def make_translation(self, sources=()):
        res = super(DataTranslator, self).make_translation(sources)
        for s in sources:
            res.source(s)
        return res


class PersonDataTranslator(BaseDataTranslator):
    """
    A person who was responsible for carrying out the translation of a data source
    manually
    """

    class_context = BASE_CONTEXT

    person = ObjectProperty(multiple=True,
            __doc__='A person responsible for carrying out the translation.')

    # No translate impl is provided here since this is intended purely as a descriptive object


def transform(transformer,
              output_key=None, output_identifier=None,
              data_sources=(), named_data_sources=None):
    """
    Do a translation with the named translator and inputs

    Parameters
    ----------
    transformer : DataTransformer
        transformer to execute
    output_key : str
        Output key. Used for generating the output's identifier. Exclusive with
        output_identifier
    output_identifier : str
        Output identifier. Exclusive with output_key
    data_sources : list of DataSource
        Input data sources
    named_data_sources : dict
        Named input data sources

    Raises
    ------
    NoTranslatorFound
        when a translator is not found
    NoSourceFound
        when a source cannot be looked up in the given context
    ExtraSourceFound
        when a more than one source is found in the given context for the given source
        identifier
    """
    if named_data_sources is None:
        named_data_sources = dict()

    if transformer is None:
        raise TypeError('No translator given')

    positional_sources = []
    for idx, psrc in enumerate(data_sources):
        if psrc is None:
            raise NoSourceFound(f'No source at position {idx}')
        loaded_src = None
        for m in psrc.load():
            if loaded_src is not None:
                raise ExtraSourceFound(f'Found more than one source for {psrc}: {loaded_src} AND {m}')
            loaded_src = m
        if loaded_src is None:
            raise NoSourceFound(f'Unable to load source at position {idx} for {psrc}')
        positional_sources.append(loaded_src)

    named_sources = dict()
    for key, nsrc in named_data_sources.items():
        if nsrc is None:
            raise NoSourceFound(f'No source for {key}')
        named_sources[key] = nsrc

    return transformer(*positional_sources,
            output_identifier=output_identifier,
            output_key=output_key,
            **named_sources)


def _lookup_translator(ctx, tname):
    for x in ctx(DataTranslator)(ident=tname).load():
        return x


def _lookup_source(ctx, sname):
    for x in ctx(DataSource)(ident=sname).load():
        return x


class NoTranslatorFound(Exception):
    '''
    Raised by `transform` when a translator cannot be found in the current context
    '''


class NoSourceFound(Exception):
    '''
    Raised by `transform` when a source cannot be found in the current context
    '''


class ExtraSourceFound(Exception):
    '''
    Raised by `transform` when more than one source is found in the current context
    '''



'''
Defines 'capabilities', pieces of functionality that an object needs which must
be injected. The receiver of the capability is called a :dfn:`capable`.

A given capability can be provided by more than one capability provider, but,
for a given set of providers, only one will be bound at a time. Logically, each
provider that provides the capability is asked, in a user-provided preference
order, whether it can provide the capability for the *specific* capable and the
first one which can provide the capability is bound to the object.

The core idea is dependency injection: a capability does not modify the capable:
the capable receives the provider and a reference to the capability provided,
but how the capable uses the provider is up to the capable. This is important
because the user of the capable should not condition its behavior on the
particular capability provider used, although it may change its behavior based
on which capabilities the capable has.

Note, that there may be some providers that lose their ability to provide a
capability after they have been bound to a capable. This loss should be
communicated with a `CannotProvideCapability` exception when the relevant
methods are called on the provider. This *may* allow certain operations to be
retried with a provider lower on the capability order, *but* a provider that
throws `CannotProvideCapability` may validly be asked if it can provide the
capability again -- if it *still* cannot provide the capability, it should
communicate that by returning `None` from its `provides_to` method.

Providers may keep state between calls to provide a capability but their
correctness must not depend on any ordering of method calls except that, of
course, their ``__init__`` is called first. For instance, a provider can retain
an index that it downloads to answer `provides_to`, but if that index can
expire, the provider should check for that and retrieve an updated index if
necessary.
'''
import six
from .utils import FCN
from itertools import chain, repeat


class _Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            _Singleton._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return _Singleton._instances[cls]


class Capability(six.with_metaclass(_Singleton)):
    '''
    A capability.
    '''
    def __init__(self):
        self._class_name = FCN(type(self))

    def __str__(self):
        return self._class_name

    def __lt__(self, other):
        if isinstance(other, Capability):
            return self._class_name < other._class_name
        return NotImplemented


class ProviderMeta(type):
    def __init__(self, name, bases, dct):
        super(ProviderMeta, self).__init__(name, bases, dct)
        my_caps = set(self.provided_capabilities)
        for base in bases:
            for base_cap in base.provided_capabilities:
                if base_cap not in my_caps:
                    my_caps.add(base_cap)
        self.provided_capabilities = sorted(my_caps)


class Provider(metaclass=ProviderMeta):
    '''
    A capability provider.

    In general, providers should do any general setup in their initializer, and setup for
    any source passed into `provides_to` method if, in fact, the provider does provide the
    needed capabilities
    '''
    provided_capabilities = []

    def provides(self, cap, obj):
        '''
        Returns a provider of the given capability if it's one this provider provides;
        otherwise, returns None.

        Parameters
        ----------
        cap : Capability
            The capability to provide
        obj : Capable
            The object to provide the capability to

        Returns
        -------
        Provider or None
        '''
        if cap in getattr(self, 'provided_capabilities', ()):
            return self.provides_to(obj, cap)

        return None

    def provides_to(self, obj, cap):
        '''
        Returns a `Provider` if the provider provides a capability to the given object;
        otherwise, returns `None`.

        The default implementation always returns `None`. Implementers of `Provider`
        should check they can actually provide the capability for the given object rather
        than just that they *might* be able to.

        It's best to do setup for providing the capability before exiting this method
        rather than, for instance, in the methods of the returned provider when the
        `Capable` is trying to use it.

        Parameters
        ----------
        obj : Capable
            The object needing/wanting the capability
        cap : Capability
            The capability needed/wanted

        Returns
        -------
        Provider or None
        '''
        return None


class Capable(object):
    '''
    An object which can have capabilities
    '''

    @property
    def needed_capabilities(self):
        '''
        The list of needed capabilities. These should be treated as though they are
        required for any of the object's methods.
        '''
        return []

    @property
    def wanted_capabilities(self):
        '''
        The list of wanted capabilities. These should be treated as though they are
        optional. The `Capable` subclass must determine how to deal with the provider not
        being available.
        '''
        return []

    def accept_capability_provider(self, cap, provider):
        '''
        The `Capable` should replace any previously accepted provider with the one
        given.

        The capability *should* be checked to determine which capability is being
        provided, even if only one is declared on the class, so that if a sub-class
        defines a capability without defining how to accept it, then the wrong actions
        won't be taken. In case the capability isn't recognized, it is generally better to
        pass it to the super() implementation rather than failing to allow for
        `cooperative multiple inheritance`__.

        __ https://rhettinger.wordpress.com/2011/05/26/super-considered-super/

        Parameters
        ----------
        cap : Capability
            the capabiilty
        provider : Provider
            the provider which provides `cap`
        '''
        raise NotImplementedError()


class CannotProvideCapability(Exception):
    '''
    Thrown by a *provider* when it cannot provide the capability during the
    object's execution
    '''
    def __init__(self, cap, provider):
        '''
        Parameters
        ----------
        cap : Capability
            the capabiilty
        provider : Provider
            the provider which failed to provide `cap`
        '''
        super(CannotProvideCapability, self).__init__('Provider, {}, cannot, now, provide the capability, {}'
                                                      .format(provider, cap))
        self._cap = cap
        self._provider = provider


class NoProviderAvailable(Exception):
    '''
    Thrown when there is no provider available for a capabiilty

    Attributes
    ----------
    cap : Capability
        The capability that was sought
    receiver : Capable
        The object for which the capability was sought
    '''
    def __init__(self, cap, receiver, providers):
        '''
        Parameters
        ----------
        cap : Capability
            The capability that was sought
        receiver : Capable
            The object for which the capability was sought
        providers : list of Provider
            Providers that were tried for the capability
        '''
        super(NoProviderAvailable, self).__init__(
                f'No providers currently provide {cap} for {receiver} among {providers}')
        self._cap = cap
        self._receiver = receiver

    @property
    def capability(self):
        return self._cap

    @property
    def receiver(self):
        return self._receiver


class NoProviderGiven(Exception):
    '''
    Thrown by a `Capable` when a `Capability` is needed, but none has been
    provided by a call to `~Capable.accept_capability_provider`
    '''
    def __init__(self, cap, receiver=None):
        '''
        Parameters
        ----------
        cap : Capability
            The capability that was sought
        receiver : Capable
            The object for which a capability was needed
        '''
        super(NoProviderGiven, self).__init__('No {} providers were given{}'
                .format(cap, ' to ' + repr(receiver) if receiver else ''))
        self._cap = cap


class UnwantedCapability(Exception):
    '''
    Thrown by a `Capable` when `~Capable.accept_capability_provider` is offered a provider
    for a capability that it does not "want", meaning it doesn't have the code to use it.
    This can happen when a sub-class of a Capable declares a needed capability without
    overriding `~Capable.accept_capability_provider` to accept that capability.
    '''


def provide(ob, provs):
    '''
    Provide capabilities to `ob` out of `provs`

    Parameters
    ----------
    ob : object
        An object which may need capabilities
    provs : list of Provider
        The providers available

    Raises
    ------
    NoProviderAvailable
        when there is no provider available
    '''
    if is_capable(ob):
        for required, cap in chain(
                zip(repeat(True), ob.needed_capabilities),
                zip(repeat(False), ob.wanted_capabilities)):
            provider = get_provider(ob, cap, provs)
            if not provider and required:
                raise NoProviderAvailable(cap, ob, provs)
            if provider:
                ob.accept_capability_provider(cap, provider)


def get_provider(ob, cap, provs):
    '''
    Get provider for a capabilty that can provide to the given object

    Parameters
    ----------
    ob : Capable
        Object needing the capability
    cap : Capability
        Capability needed
    provs : list of Provider
        All providers available

    Returns
    -------
    Provider
        A provider of the given capability or `None`
    '''
    for provider in get_providers(cap, provs, ob):
        return provider
    return None


def get_providers(cap, provs, ob):
    '''
    Get providers for a capabilty

    Parameters
    ----------
    cap : Capability
        Capability needed
    provs : list of Provider
        All providers available

    Yields
    ------
    Provider
        A Provider that provides the given capability
    '''
    for p in provs:
        provfn = p.provides(cap, ob)
        if provfn:
            yield provfn


def is_capable(ob):
    '''
    Returns true if the given object can accept capability providers

    Parameters
    ----------
    ob : object
        An object which may be a `Capable`

    Returns
    -------
    bool
        True if the given object accepts capability providers of some kind. Otherwise,
        false.
    '''
    return isinstance(ob, Capable)



'''
Classes for managing things in the owmeta-core project directory, typically named .owm
'''
import hashlib
from os import makedirs, getcwd, rename
from os.path import join as pth_join, exists
from random import getrandbits
from shutil import rmtree
import logging
from tempfile import mkdtemp

from .capabilities import (OutputFilePathProvider,
        FilePathProvider,
        FilePathCapability,
        CacheDirectoryProvider,
        TemporaryDirectoryProvider)
from .datasource import DataSource
from .file_lock import lock_file
from .utils import FCN


L = logging.getLogger(__name__)


class WorkingDirectoryProvider(FilePathProvider):
    '''
    Provides file paths from the current working directory for
    `.data_trans.local_file_ds.LocalFileDataSource` instances.
    '''

    def __init__(self, cwd=None, **kwargs):
        '''
        Parameters
        ----------
        cwd : str or pathlib.Path, optional
            The working directory to use. The default is what `os.getcwd` returns
        '''
        super().__init__(**kwargs)
        if cwd is not None:
            self.cwd = cwd
        else:
            self.cwd = getcwd()

    def provides_to(self, obj, cap):
        from .data_trans.local_file_ds import LocalFileDataSource
        file_name = obj.file_name.one()
        if not file_name:
            return None
        if (isinstance(obj, LocalFileDataSource) and
                exists(pth_join(self.cwd, file_name))):
            return self
        return None

    def file_path(self):
        return self.cwd


class TransactionalDataSourceDirProvider(OutputFilePathProvider, FilePathProvider):
    '''
    Provides a DataSourceDirectoryProvider with transactional semantics.

    Provides a `TDSDPHelper` for `.DataSource` objects, indexed by the `DataSource`
    identifier. If asked to provide a `.FilePathCapability` (i.e, a directory for input),
    and the `DataSource` is a `~.data_trans.local_file_ds.LocalFileDataSource`, then we'll
    check that a file named with the value of `~.LocalFileDataSource.file_name` is in the
    provided directory.
    '''
    def __init__(self, basedir, transaction_manager):
        self._basedir = basedir
        self.transaction_manager = transaction_manager
        self._providers = dict()

    def provides_to(self, obj, cap):
        from .data_trans.local_file_ds import LocalFileDataSource
        if isinstance(obj, DataSource):
            key = hashlib.sha256(obj.identifier.encode('utf-8')).hexdigest()
            provider = self._providers.get(key)
            if provider is None:
                provider = TDSDPHelper(self._basedir, key, self.transaction_manager)
            if cap == FilePathCapability():
                # We need the file to already exist for FilePathCapability, so here we
                # check for that
                provider_file_path = provider.file_path()
                if not exists(provider_file_path):
                    return None
                if isinstance(obj, LocalFileDataSource):
                    file_name = obj.file_name.one()
                    if not exists(pth_join(provider_file_path, file_name)):
                        return None
            self._providers[key] = provider
            return provider
        return None


class TDSDPHelper(FilePathProvider, OutputFilePathProvider):
    '''
    This provider relies on the `transaction` library's machinery to manage the
    transaction.

    Consistency is NOT guaranteed in all cases: in particular, this provider uses a
    file-based locking mechanism with a `"lock file" <.file_lock.lock_file>` in the given
    base directory which, if it's deleted during the two-phase commit process, removes the
    isolation of the changes made in the directory.
    '''

    def __init__(self, basedir, key, transaction_manager):
        self._basedir = basedir
        self._key = key
        self.transaction_manager = transaction_manager

        self._uncommitted_path = None
        self._dm_sort_key = None
        self._file_lock = None
        self._committed_path = pth_join(self._basedir, self._key)
        self._prev_version_path = f"{self._committed_path}.prev"
        self._lock_file_name = f"{self._committed_path}.lock"
        self._transaction = None

    def file_path(self):
        if self._uncommitted_path is not None:
            return self._uncommitted_path
        else:
            return self._committed_path

    def output_file_path(self):
        if self._uncommitted_path is None:
            for attempt in range(3):
                try:
                    random_suffix = getrandbits(32)
                    path = f"{self._committed_path}.uncommitted.{random_suffix}"
                    makedirs(path)
                except OSError:
                    pass
                else:
                    break
            else: # no break
                raise Exception('Unable to create output directory')

            self._uncommitted_path = path
            # Join the transaction so if it fails our abort method will get called so
            # we can delete the "uncommitted" directory
            self._file_lock = lock_file(self._lock_file_name, unique_key=random_suffix.to_bytes(4, 'little'))
            self._transaction = self.transaction_manager.get()
            self._transaction.join(self)
        return self._uncommitted_path

    def tpc_begin(self, transaction):
        self._transaction = transaction

    def commit(self, transaction):
        self._file_lock.acquire()
        if exists(self._committed_path):
            rename(self._committed_path, self._prev_version_path)
        rename(self._uncommitted_path, self._committed_path)
        self._uncommitted_path = None

    def tpc_vote(self, transaction):
        if transaction is not self._transaction:
            raise Exception('tcp_vote called with the wrong transaction: given'
                    f' {transaction}, but expected {self._transaction}')

    def tpc_abort(self, transaction):
        try:
            if exists(self._committed_path):
                rmtree(self._committed_path)
            if exists(self._prev_version_path):
                rename(self._prev_version_path, self._committed_path)
        except Exception:
            L.error('Received exception in tpc_abort for transaction for %s,'
                    ' back from %s',
                    self._committed_path, self._prev_version_path,
                    exc_info=True)
        finally:
            self._release_lock()

    def abort(self, transaction):
        rmtree(self._uncommitted_path)
        self._uncommitted_path = None

    def _handle_rmtree_error(self, retry, path, exc_info):
        # abort and finish should "never fail" so we just log errors
        L.error('Unable to delete %s for transaction commit/abort', path,
                exc_info=exc_info)

    def tpc_finish(self, transaction):
        try:
            if exists(self._prev_version_path):
                rmtree(self._prev_version_path, onerror=self._handle_rmtree_error)
        except Exception:
            L.error('Received exception in tpc_finish for transaction for %s,'
                    ' removing %s,',
                    self._committed_path, self._prev_version_path,
                    exc_info=True)
        finally:
            self._release_lock()

    def _release_lock(self):
        try:
            self._file_lock.release()
        except FileNotFoundError:
            L.error('Lock file was deleted before being released: directory contents may be'
                    ' inconsistent', exc_info=True)
        except PermissionError:
            L.error('Lock file could not be released due to a permissions error: correct'
                    ' file system permissions, check directory contents, and delete the'
                    ' lock file.', exc_info=True)
        except Exception:
            L.error('Unknown error during lock file release: directory contents may be '
                    ' inconsistent', exc_info=True)

    def sortKey(self):
        '''
        See Also
        --------
        transaction.interfaces.IDataManager
        '''
        if self._dm_sort_key is None:
            if self._uncommitted_path is None:
                raise Exception('output_file_path must be called before a sort key is created')
            cname = FCN(type(self))
            self._dm_sort_key = f'{cname}{self._uncommitted_path}'
        return self._dm_sort_key


class SimpleDataSourceDirProvider(OutputFilePathProvider):
    '''
    Provides a directory under the provided base directory
    '''

    def __init__(self, basedir):
        self._basedir = basedir

    def provides_to(self, obj, cap):
        if isinstance(obj, DataSource):
            key = hashlib.sha256(obj.identifier.encode('utf-8')).hexdigest()
            return type(self).Helper(self, key)
        return None

    class Helper(OutputFilePathProvider):
        def __init__(self, parent, key):
            self._key = key
            self._parent = parent

        def output_file_path(self):
            # When a file path is requested, then we create one?
            #   Well, if we try to create the file path here, but that fails, there's not
            #   really a good recourse: we could have tried a different provider earlier in
            #
            #   provides_to, but not now...  If, instead, we do it in provides_to, then we can
            #   just log the error and say "no, I can't provide a directory"
            #
            # An important point about the mapping from the RDF graph:
            #   We're mapping from the RDF graph here to a single directory, so what we're
            #   implicitly saying is that there's one and only one set of files corresponding
            #   to the data source. For a work-in-progress project, this is really
            #   important...also the data could change over time, right? How would we deal
            #   with a different version of the data still pointing to the project directory?
            #   We aren't worried about that because everything in the project is WIP. What
            #   you would worry about is whether you could have multiple WIP versions for a
            #   given data source...that's not as clear cut to me
            #
            # create a directory
            # create an index entry
            #   the index stores the mapping to the directory
            #
            # what if the directory creation fails?
            #   Then we won't have a directory and we won't create the index entry
            # what if the index entry creation fails?
            #   Then we need to clean delete the directory before propagating the exception
            #       What if the directory clean up fails?
            #           Then we need to report that as well as the original exception --
            #           Python will handle that for us though.
            return pth_join(self._parent._basedir, self._key)


class SimpleCacheDirectoryProvider(CacheDirectoryProvider):
    '''
    Provides a directory for caching remote resources as local files
    '''

    def __init__(self, cache_directory, **kwargs):
        super().__init__(**kwargs)
        self._cache_directory = cache_directory

    def provides_to(self, obj, cap):
        return self

    def cache_directory(self, cache_key):
        res = pth_join(self._cache_directory, cache_key)
        makedirs(res, exist_ok=True)
        return res


class SimpleTemporaryDirectoryProvider(TemporaryDirectoryProvider):
    '''
    Provides temporary directories under a given base directory
    '''
    def __init__(self, base_directory, suffix=None, prefix=None, **kwargs):
        super().__init__(**kwargs)
        self._base_directory = base_directory
        self._suffix = suffix
        self._prefix = prefix

    def provides_to(self, obj, cap):
        return self

    def temporary_directory(self):
        if not exists(self._base_directory):
            makedirs(self._base_directory)
        return mkdtemp(suffix=self._suffix, prefix=self._prefix, dir=self._base_directory)



from contextlib import contextmanager

from .. import BASE_CONTEXT
from ..datasource import Informational, DataSource


class FileDataSource(DataSource):
    '''
    This DataSource represents a "file", essentially a sequence of bytes with a name

    Attributes
    ----------
    source_file_path : :term:`path-like object`
        The file to commit for this datasource
    '''

    class_context = BASE_CONTEXT

    md5 = Informational(display_name='MD5 hash')
    sha256 = Informational(display_name='SHA-256 hash')
    sha512 = Informational(display_name='SHA-512 hash')

    def __init__(self, *args, **kwargs):
        super(FileDataSource, self).__init__(*args, **kwargs)
        self.source_file_path = None

    @contextmanager
    def file_contents(self):
        '''
        Returns a :term:`file object` for reading data from the file
        '''
        raise NotImplementedError()

    def update_hash(self, algorithm):
        '''
        Set a message digest property for the file

        Parameters
        ----------
        algorithm : str
            The name of the property and algorithm to update
        '''
        import hashlib
        hsh = hashlib.new(algorithm)
        with self.file_contents() as f:
            hsh.update(f.read())
        getattr(self, algorithm).set(hsh.hexdigest())



from __future__ import print_function

from rdflib.graph import Graph
from rdflib.term import Literal, URIRef
from rdflib.store import Store

# Directions for traversal across triples
UP = 'up'
''' Object to Subject direction for traversal across triples. '''

DOWN = 'down'
''' Subject to Object direction for traversal across triples. '''


def print_graph(g, hide_namespaces=False):
    s = g.serialize(format='n3').decode("UTF-8")
    if hide_namespaces:
        lines = s.splitlines()
        s = "\n".join(l for l in lines if not l.startswith("@prefix"))
    print(s)


def serialize_rdflib_term(x, namespace_manager=None):
    return x.n3(namespace_manager)


def deserialize_rdflib_term(x):
    if isinstance(x, Literal):
        x = x.toPython()
        if isinstance(x, Literal):
            x = str(x)
    return x


def triple_to_n3(trip, namespace_manager=None):
    p = ''
    ns = set([])
    for x in trip:
        s = serialize_rdflib_term(x, namespace_manager)
        if isinstance(x, URIRef) and s[0] != '<':
            ns.add(s.split(':', 1)[0])
        elif isinstance(x, Literal) and '^^' in s and s[-1] != '>':
            ns.add(s.split('^^', 1)[1].split(':', 1)[0])

        p += s + ' '
    return p


def triples_to_bgp(trips, namespace_manager=None, show_namespaces=False):
    # XXX: Collisions could result between the variable names of different
    # objects
    g = ""
    ns = set([])
    for y in trips:
        g += triple_to_n3(y) + ".\n"

    if (namespace_manager is not None) and show_namespaces:
        g = "".join('@prefix ' + str(x) + ': ' + y.n3() + ' .\n'
                    for x, y
                    in namespace_manager.namespaces()
                    if x in ns) + g

    return g


_none_singleton_set = frozenset([None])


def transitive_lookup(graph, start, predicate, context=None, direction=DOWN, seen=None):
    '''
    Do a transitive lookup over an `rdflib.graph.Graph` or `rdflib.store.Store`

    In other words, finds all resources which relate to `start` through zero or more
    `predicate` relationships. `start` itself will be included in the return value.

    Loops in the input `graph` will not cause non-termination.

    Parameters
    ----------
    graph : rdflib.graph.Graph or rdflib.store.Store
        The graph to query
    start : rdflib.term.Identifier
        The resource in the graph to start from
    predicate : rdflib.term.URIRef
        The predicate relating terms in the closure
    context : rdflib.graph.Graph or rdflib.term.URIRef
        The context in which the query should run. Optional
    direction : DOWN or UP
        The direction in which to traverse
    seen : set of rdflib.term.Identifier
        A set of terms which have already been "seen" by the algorithm. Useful for
        repeated calls to `transitive_lookup`. Note: if the `start` is in `seen`, queries
        from `start` will still be done, but any items in the result of *those* queries
        will not be queried for if in `seen`. Optional

    Returns
    -------
    set of rdflib.term.Identifier
        resources in the transitive closure of `predicate` from `start`
    '''
    if seen:
        res = seen
    else:
        res = set()
    border = set([start])
    while border:
        new_border = set()
        if direction is DOWN:
            qx = (list(border), predicate, None)
            idx = 2
        else:
            qx = (None, predicate, list(border))
            idx = 0
        for t in graph.triples_choices(qx, context=context):
            if isinstance(t[0], tuple):
                o = t[0][idx]
            else:
                o = t[idx]
            if o not in res:
                new_border.add(o)
        res |= border
        border = new_border
    res -= _none_singleton_set
    return res


class BatchAddGraph(object):
    ''' Wrapper around graph that turns calls to 'add' into calls to 'addN' '''
    def __init__(self, graph, batchsize=1000, _parent=None, *args, **kwargs):
        self.graph = graph
        self.g = (graph,)
        if _parent:
            self.batch = _parent.batch
            self.batchsize = _parent.batchsize
            self._parent = _parent
        else:
            self.batchsize = batchsize
            self._parent = None
            self.reset()

    def reset(self):
        self.batch = []
        self._count = 0

    @property
    def count(self):
        if self._parent:
            return self._parent.count
        else:
            return self._count

    @count.setter
    def count(self, value):
        if self._parent:
            self._parent.count = value
        else:
            self._count = value

    def add(self, triple):
        if self.count > 0 and self.count % self.batchsize == 0:
            self.graph.addN(self.batch)
            self.batch = []
        self.count += 1
        self.batch.append(triple + self.g)

    def get_context(self, ctx):
        return BatchAddGraph(self.graph.get_context(ctx), _parent=self)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, *exc):
        if exc[0] is None:
            self.graph.addN(self.batch)


transitive_subjects = transitive_lookup
''' Alias to `transitive_lookup` '''


class ContextSubsetStore(Store):
    # Returns triples imported by the given context
    context_aware = True

    def __init__(self, store, **kwargs):
        super(ContextSubsetStore, self).__init__(**kwargs)
        self.__store = store
        self.__context_ids = None

    def init_contexts(self):
        raise NotImplementedError

    def __init_contexts(self):
        if self.__context_ids is None:
            self.__context_ids = self.init_contexts()

    def triples(self, pattern, context=None):
        self.__init_contexts()

        ctx = self._determine_context(context)
        if ctx is _BAD_CONTEXT:
            return

        # If the sum of lengths of the selected contexts is less than total number of
        # triples, query each context in series
        for t in self.__store.triples(pattern, ctx):
            contexts = set(getattr(c, 'identifier', c) for c in t[1])
            if self.__context_ids:
                inter = self.__context_ids & contexts
            else:
                inter = contexts
            if inter:
                yield t[0], inter

    def remove(self, pattern, context=None):
        self.__init_contexts()

        ctx = self._determine_context(context)
        if ctx is _BAD_CONTEXT:
            return
        for t in self.__store.triples(pattern, ctx):
            triple = t[0]
            contexts = set(getattr(c, 'identifier', c) for c in t[1])
            if self.__context_ids:
                inter = self.__context_ids & contexts
            else:
                inter = contexts
            for ctx in inter:
                self.__store.remove((triple[0], triple[1], triple[2]), ctx)

    def triples_choices(self, pattern, context=None):
        self.__init_contexts()

        ctx = self._determine_context(context)
        if ctx is _BAD_CONTEXT:
            return

        for t in self.__store.triples_choices(pattern, ctx):
            contexts = set(getattr(c, 'identifier', c) for c in t[1])
            if self.__context_ids:
                inter = self.__context_ids & contexts
            else:
                inter = contexts

            if inter:
                yield t[0], inter

    def _determine_context(self, context):
        # This is a method that has to contend with RDFLib's abiding confusion over
        # whether Store's should return Graphs. This is stupid, because of course they
        # shouldn't, but RDFLib acts like they should...and so here we are
        context_id = getattr(context, 'identifier', context)
        if context_id is not None and context_id not in self.__context_ids:
            return _BAD_CONTEXT
        if len(self.__context_ids) == 1 and context_id is None:
            # Micro-benchmarked this with timeit: it's faster than tuple(s)[0] and
            # next(iter(s),None)
            for context_id in self.__context_ids:
                break
        if context_id is None:
            return None
        # We shouldn't be querying from this store, but we pass in ourselves as the store
        # to prevent RDFLib from making a new memory story
        return Graph(identifier=context_id, store=self)

    def contexts(self, triple=None):
        if triple is not None:
            for x in self.triples(triple):
                for c in x[1]:
                    yield getattr(c, 'identifier', c)
        else:
            self.__init_contexts()
            for c in self.__context_ids:
                yield c

    def namespace(self, prefix):
        return self.__store.namespace(prefix)

    def prefix(self, uri):
        return self.__store.prefix(uri)

    def bind(self, prefix, namespace, override=True):
        return self.__store.bind(prefix, namespace, override=override)

    def namespaces(self):
        for x in self.__store.namespaces():
            yield x

    def __str__(self):
        return f'{type(self).__name__}(store={self.__store})'


_BAD_CONTEXT = object()
